"""Parameters of the IOCs of the AP discipline."""
from copy import deepcopy as _dcopy

_off = 0
_on = 1


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.


_si_bbb_frontbackend = [
    ['SI-Glob:DI-BbBProc-L:FBE_BE_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_Z_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_X_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_Y_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBE_BE_PHASE', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_X_PHASE_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_Y_PHASE_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_MODE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_SIGN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_GAIN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_OFFSET', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_X_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_Y_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_FAN_MODE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_FAN_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FBELT_SERVO_MAXDELTA', 0.0, 0.0],

    ['SI-Glob:DI-BbBProc-V:FBE_BE_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_Z_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_X_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_Y_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBE_BE_PHASE', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_X_PHASE_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_Y_PHASE_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_MODE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_SIGN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_GAIN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_OFFSET', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_X_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_Y_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_FAN_MODE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_FAN_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FBELT_SERVO_MAXDELTA', 0.0, 0.0],

    ['SI-Glob:DI-BbBProc-H:FBE_BE_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_Z_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_X_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_Y_ATT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBE_BE_PHASE', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_X_PHASE_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_Y_PHASE_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_MODE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_SIGN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_GAIN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_OFFSET', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_X_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_Y_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_FAN_MODE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_FAN_SETPT', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FBELT_SERVO_MAXDELTA', 0.0, 0.0],
    ]

_si_bbb_l = [
    ['SI-Glob:DI-BbBProc-L:SETSEL', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:SHIFTGAIN', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:PROC_DS', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:SAT_THRESHOLD', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:GDEN', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:CSET0', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-L:CSET1', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-L:CSET2', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-L:CSET3', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-L:FB_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-L:CF_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-L:FB_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-L:CF_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE0_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE1_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE2_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE0_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE1_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE2_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-L:PHTRK_LOOPCTRL', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:DRIVE2_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:PHTRK_MOD', 0, 0.0],
    ['SI-Glob:DI-BbBProc-L:TADC', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:TDAC', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:DELAY', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:OFF_FIDS', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:FID_DELAY', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:ECLDEL0', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:ECLDEL1', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:ECLDEL2', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:ECLDEL3', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_0_FWDLOSS', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_0_REVLOSS', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_1_FWDLOSS', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-L:MCLRAW_1_REVLOSS', 0.0, 0.0],
    ]

_si_bbb_h = [
    ['SI-Glob:DI-BbBProc-H:SETSEL', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:SHIFTGAIN', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:PROC_DS', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:SAT_THRESHOLD', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:GDEN', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:CSET0', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-H:CSET1', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-H:CSET2', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-H:CSET3', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-H:FB_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-H:CF_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-H:FB_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-H:CF_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_ENABLE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_PERIOD', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_SPAN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_TUNE', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:CLEAN_AMPL', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE0_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE1_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE2_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE0_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE1_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE2_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-H:PHTRK_LOOPCTRL', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:DRIVE2_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:PHTRK_MOD', 0, 0.0],
    ['SI-Glob:DI-BbBProc-H:TADC', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:TDAC', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:DELAY', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:OFF_FIDS', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:FID_DELAY', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:ECLDEL0', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:ECLDEL1', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:ECLDEL2', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-H:ECLDEL3', 0.0, 0.0],
    ['SI-Glob:DI-BbBAmpH:Enbl-Sel', 0, 0.0],
    ['SI-Glob:DI-BbBAmpH:GainAuto-Sel', 0, 0.0],
    ['SI-Glob:DI-BbBAmpH:Gain-SP', 0.0, 0.0],
    ['SI-Glob:DI-BbBAmpH:GainStep-SP', 0.0, 0.0],
    ]

_si_bbb_v = [
    ['SI-Glob:DI-BbBProc-V:SETSEL', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:SHIFTGAIN', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:PROC_DS', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:SAT_THRESHOLD', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:GDEN', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:CSET0', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-V:CSET1', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-V:CSET2', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-V:CSET3', [0.0]*31, 0.0],
    ['SI-Glob:DI-BbBProc-V:FB_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-V:CF_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-V:FB_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-V:CF_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_ENABLE', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_PERIOD', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_SPAN', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_TUNE', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:CLEAN_AMPL', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE0_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE1_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE2_MASK', [0]*864, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE0_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE1_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE2_PATTERN', '', 0.0],
    ['SI-Glob:DI-BbBProc-V:PHTRK_LOOPCTRL', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:DRIVE2_TRACK', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:PHTRK_MOD', 0, 0.0],
    ['SI-Glob:DI-BbBProc-V:TADC', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:TDAC', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:DELAY', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:OFF_FIDS', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:FID_DELAY', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:ECLDEL0', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:ECLDEL1', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:ECLDEL2', 0.0, 0.0],
    ['SI-Glob:DI-BbBProc-V:ECLDEL3', 0.0, 0.0],
    ['SI-Glob:DI-BbBAmpV:Enbl-Sel', 0, 0.0],
    ['SI-Glob:DI-BbBAmpV:GainAuto-Sel', 0, 0.0],
    ['SI-Glob:DI-BbBAmpV:Gain-SP', 0.0, 0.0],
    ['SI-Glob:DI-BbBAmpV:GainStep-SP', 0.0, 0.0],
    ]

_si_sofb = [
    ['SI-Glob:AP-SOFB:SOFBMode-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:RefOrbX-SP', [0.0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:RefOrbY-SP', [0.0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:SmoothNrPts-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:OrbAcqRate-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:SmoothMethod-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:LoopFreq-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopMaxOrbDistortion-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKpCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKpCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKpRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKiCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKiCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKiRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKdCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKdCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:LoopPIDKdRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxKickCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxKickCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxDeltaKickCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxDeltaKickCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxDeltaKickRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:CorrSync-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:CorrPSSOFBWait-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:CorrPSSOFBEnbl-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:FOFBDownloadKicks-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:FOFBUpdateRefOrb-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:FOFBNullSpaceProj-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:FOFBZeroDistortionAtBPMs-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:FOFBDownloadKicksPerc-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:FOFBUpdateRefOrbPerc-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:RespMat-SP', [0.0, ]*449600, 0.0],
    ['SI-Glob:AP-SOFB:RespMatMode-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:BPMXEnblList-SP', [0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:BPMYEnblList-SP', [0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:CHEnblList-SP', [0, ]*120, 0.0],
    ['SI-Glob:AP-SOFB:CVEnblList-SP', [0, ]*160, 0.0],
    ['SI-Glob:AP-SOFB:RFEnbl-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:MinSingValue-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:TikhonovRegConst-SP', 0.0, 0.0],
    ]


_si_fofb = [
    ['SI-Glob:AP-FOFB:LoopGainH-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:LoopGainV-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:LoopMaxOrbDistortion-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:LoopMaxOrbDistortionEnbl-Sel', 0, 0.0],
    ['SI-Glob:AP-FOFB:LoopPacketLossDetecEnbl-Sel', 0, 0.0],
    ['SI-Glob:AP-FOFB:CorrSetCurrZeroDuration-SP', 0, 0.0],
    ['SI-Glob:AP-FOFB:CHAccSatMax-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:CVAccSatMax-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:TimeFrameLen-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:CtrlrSyncUseEnblList-Sel', 0, 0.0],
    ['SI-Glob:AP-FOFB:KickBufferSize-SP', 0, 0.0],
    ['SI-Glob:AP-FOFB:RefOrbX-SP', [0.0, ]*160, 0.0],
    ['SI-Glob:AP-FOFB:RefOrbY-SP', [0.0, ]*160, 0.0],
    ['SI-Glob:AP-FOFB:BPMXEnblList-SP', [0, ]*160, 0.0],
    ['SI-Glob:AP-FOFB:BPMYEnblList-SP', [0, ]*160, 0.0],
    ['SI-Glob:AP-FOFB:CHEnblList-SP', [0, ]*80, 0.0],
    ['SI-Glob:AP-FOFB:CVEnblList-SP', [0, ]*80, 0.0],
    ['SI-Glob:AP-FOFB:UseRF-Sel', 0, 0.0],
    ['SI-Glob:AP-FOFB:RespMat-SP', [0.0, ]*51520, 0.0],
    ['SI-Glob:AP-FOFB:MinSingValue-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:TikhonovRegConst-SP', 0.0, 0.0],
    ['SI-Glob:AP-FOFB:InvRespMatNormMode-Sel', 0, 0.0],
]


pvs = _si_sofb + _si_fofb
pvs += _si_bbb_frontbackend + _si_bbb_l + _si_bbb_h + _si_bbb_v
_template_dict = {'pvs': pvs}



"""InjSys Standby Handler devices."""

import time as _time
from copy import deepcopy as _dcopy
from threading import Thread as _Thread, Event as _Flag
import logging as _log

from .device import Devices as _Devices, DeviceNC as _DeviceNC
from .lillrf import DevLILLRF
from .modltr import LIModltr
from .pwrsupply import PowerSupply, PowerSupplyPU
from .timing import EVG, Event, Trigger
from .rf import ASLLRF
from .posang import PosAng

from ..search import PSSearch, HLTimeSearch
from ..csdev import Const as _Const
from ..timesys.csdev import Const as _TIConst
from ..pwrsupply.csdev import Const as _PSConst
from ..injctrl.csdev import Const as _InjConst
from ..callbacks import Callback as _Callback


class _BaseHandler(_Devices):
    """Base standby mode handler for injection procedure."""

    @property
    def is_on(self):
        """Is on."""
        for dev, ppties in self._on_values.items():
            for ppty, val in ppties.items():
                if not dev[ppty] == val:
                    return False
        return True

    @property
    def on_values(self):
        """Property values equivalent to 'on' status."""
        vals = dict()
        for dev, ppties in self._on_values.items():
            for ppty, val in ppties.items():
                vals[dev.pv_object(ppty).pvname] = val
        return vals


class ASPUStandbyHandler(_BaseHandler):
    """Pulsed magnets standby mode handler for injection procedure."""

    def __init__(self):
        """Init."""
        self._punames = PSSearch.get_psnames(
            {'dis': 'PU', 'dev': '.*(Kckr|Sept)'})
        self._trignames = [
            dev.replace('PU', 'TI') for dev in self._punames]
        self._modnames = LIModltr.DEVICES.ALL

        # pu devices
        self._pudevs = [PowerSupplyPU(pun) for pun in self._punames]

        # trigger devices
        self._trigdevs = [Trigger(trg) for trg in self._trignames]

        # modulator devices
        self._moddevs = [LIModltr(mod) for mod in self._modnames]
        self._limps = _DeviceNC(
            'LA-CN:H1MPS-1',
            ('Mod1State_I', 'Mod1State_L', 'Mod1State_R',
             'Mod2State_I', 'Mod2State_L', 'Mod2State_R'))

        alldevs = self._pudevs + self._trigdevs + self._moddevs
        alldevs.append(self._limps)
        alldevs = tuple(alldevs)

        self._on_values = dict()
        for pudev in self._pudevs:
            if 'InjDpKckr' in pudev.devname:
                continue
            self._on_values[pudev] = {
                'Pulse-Sts': _PSConst.DsblEnbl.Enbl,
                'PwrState-Sts': _PSConst.DsblEnbl.Enbl}
        for tdev in self._trigdevs:
            if 'InjDpKckr' in tdev.devname:
                continue
            self._on_values[tdev] = {
                'State-Sts': _TIConst.DsblEnbl.Enbl}
        for mdev in self._moddevs:
            self._on_values[mdev] = {
                'CHARGE': _TIConst.DsblEnbl.Enbl,
                'TRIGOUT': _TIConst.DsblEnbl.Enbl,
                'CPS_ALL': 1,
            }
        self._on_values[self._limps] = {
            'Mod1State_I': 0,
            'Mod1State_L': 0,
            'Mod2State_I': 0,
            'Mod2State_L': 0,
        }

        # call base class constructor
        super().__init__('', alldevs)

    @property
    def punames(self):
        """Power supply names."""
        return _dcopy(self._punames)

    @property
    def pudevices(self):
        """Power supply devices."""
        return self._pudevs

    @property
    def trignames(self):
        """Trigger names."""
        return _dcopy(self._trignames)

    @property
    def trigdevices(self):
        """Trigger devices."""
        return self._trigdevs

    @property
    def modnames(self):
        """Modulator names."""
        return _dcopy(self._modnames)

    @property
    def moddevices(self):
        """Modulator devices."""
        return self._moddevs

    def cmd_turn_off(self):
        """Turn off."""
        # turn modulator trigout off
        self._set_devices_propty(
            self._moddevs, 'TRIGOUT', _Const.DsblEnbl.Dsbl)

        # wait for modulator trigout to turn off
        retval = self._wait_devices_propty(
            self._moddevs, 'TRIGOUT', _Const.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI modulator TrigOut to be off timed '\
                   'out without success! Verify LI Modulators!'
            return [False, text, retval[1]]

        # set pulsed magnet pulse off
        self._set_devices_propty(
            self._pudevs, 'Pulse-Sel', _PSConst.DsblEnbl.Dsbl)

        # wait for pulsed magnet pulse to be off
        retval = self._wait_devices_propty(
            self._pudevs, 'Pulse-Sts', _PSConst.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for pulsed magnet Pulse to be disabled '\
                   'timed out without success! Verify pulsed magnets!'
            return [False, text, retval[1]]

        # set pulsed magnet power state off
        self._set_devices_propty(
            self._pudevs, 'PwrState-Sel', _PSConst.DsblEnbl.Dsbl)

        # wait for pulsed magnet power state to be off
        retval = self._wait_devices_propty(
            self._pudevs, 'PwrState-Sts', _PSConst.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for pulsed magnet PwrState to be off '\
                   'timed out without success! Verify pulsed magnets!'
            return [False, text, retval[1]]

        # wait for modulators trig.out
        _time.sleep(1)

        # turn modulator charge off
        self._set_devices_propty(
            self._moddevs, 'CHARGE', _Const.DsblEnbl.Dsbl)

        # wait for modulator charge to turn off
        retval = self._wait_devices_propty(
            self._moddevs, 'CHARGE', _Const.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI modulator Charge to be off timed '\
                   'out without success! Verify LI Modulators!'
            return [False, text, retval[1]]

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        # turn modulators charge on
        self._set_devices_propty(
            self._moddevs, 'CHARGE', _Const.DsblEnbl.Enbl)

        # wait for modulators charge to turn on
        retval = self._wait_devices_propty(
            self._moddevs, 'CHARGE', _Const.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI modulator Charge to be on timed '\
                   'out without success! Verify LI Modulators!'
            return [False, text, retval[1]]

        # reset modulator
        for dev in self._moddevs:
            if not dev.cmd_reset():
                text = 'Could not reset LI modulator! Verify LI Modulators!'
                return [False, text, [dev.devname+':RESET', ]]

        devs = [dev for dev in self._pudevs if 'InjDpKckr' not in dev.devname]

        # set pulsed magnet power state on
        self._set_devices_propty(devs, 'PwrState-Sel', _PSConst.DsblEnbl.Enbl)

        # wait for pulsed magnet power state to be on
        retval = self._wait_devices_propty(
            devs, 'PwrState-Sts', _PSConst.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for pulsed magnet PwrState to be on '\
                   'timed out without success! Verify pulsed magnets!'
            return [False, text, retval[1]]

        # wait a moment for the PU high voltage and modulators charge
        _time.sleep(1)

        # set pulsed magnet pulse on
        self._set_devices_propty(devs, 'Pulse-Sel', _PSConst.DsblEnbl.Enbl)

        # wait for pulsed magnet pulse to be on
        retval = self._wait_devices_propty(
            devs, 'Pulse-Sts', _PSConst.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for pulsed magnet Pulse to be enabled '\
                   'timed out without success! Verify pulsed magnets!'
            return [False, text, retval[1]]

        # turn modulator trigout on
        self._set_devices_propty(
            self._moddevs, 'TRIGOUT', _Const.DsblEnbl.Enbl)

        # wait for modulator trigout to turn on
        retval = self._wait_devices_propty(
            self._moddevs, 'TRIGOUT', _Const.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI modulator TrigOut to be on timed '\
                   'out without success! Verify LI Modulators!'
            return [False, text, retval[1]]

        # check if mps status is ok
        _t0 = _time.time()
        while _time.time() - _t0 < 5:
            if not self._limps['Mod1State_I'] and \
                    not self._limps['Mod2State_I']:
                break
        else:
            problems = [
                self._limps.pv_object('Mod'+i+'State_I').pvname for
                i in ['1', '2'] if self._limps['Mod'+i+'State_I']]
            text = 'Check for LI modulators MPS Status to be ok timed '\
                   'out without success! Verify LI Modulators MPS!'
            return [False, text, problems]

        # reset linac mps modulator signal
        self._limps['Mod1State_R'] = 1
        self._limps['Mod2State_R'] = 1
        _time.sleep(1)
        self._limps['Mod1State_R'] = 0
        self._limps['Mod2State_R'] = 0

        # check if mps latch is ok
        _t0 = _time.time()
        while _time.time() - _t0 < 5:
            if not self._limps['Mod1State_L'] and \
                    not self._limps['Mod2State_L']:
                break
        else:
            problems = [
                self._limps.pv_object('Mod'+i+'State_L').pvname for
                i in ['1', '2'] if self._limps['Mod'+i+'State_L']]
            text = 'Check for LI modulators MPS Latch to be ok timed '\
                   'out without success! Verify LI Modulators MPS!'
            return [False, text, problems]

        return True, '', []


class BOPSRampStandbyHandler(_BaseHandler):
    """BO PS Ramp standy mode handler for injection procedure."""

    def __init__(self):
        """Init."""
        self._psnames = PSSearch.get_psnames(
            {'sec': 'BO', 'dis': 'PS'})
        self._trignames = HLTimeSearch.get_hl_triggers(
            {'sec': 'BO', 'dev': 'Mags'})

        # ps devices
        self._psdevs = [PowerSupply(psn) for psn in self._psnames]

        # trigger devices
        self._trigdevs = [Trigger(trg) for trg in self._trignames]

        alldevs = tuple(self._psdevs + self._trigdevs)

        self._on_values = dict()
        for psdev in self._psdevs:
            self._on_values[psdev] = {
                'OpMode-Sts': _PSConst.States.RmpWfm,
                'WfmUpdateAuto-Sts': _PSConst.DsblEnbl.Enbl}
        for tdev in self._trigdevs:
            self._on_values[tdev] = {
                'State-Sts': _TIConst.DsblEnbl.Enbl,
                'Src-Sts': tdev.source_options.index('RmpBO')}

        # call base class constructor
        super().__init__('', alldevs)

    @property
    def psnames(self):
        """Power supply names."""
        return _dcopy(self._psnames)

    @property
    def psdevices(self):
        """Power supply devices."""
        return self._psdevs

    @property
    def trignames(self):
        """Trigger names."""
        return _dcopy(self._trignames)

    @property
    def trigdevices(self):
        """Trigger devices."""
        return self._trigdevs

    def cmd_turn_off(self):
        """Turn off."""
        # disable triggers
        self._set_devices_propty(
            self._trigdevs, 'State-Sel', _TIConst.DsblEnbl.Dsbl)

        # wait for triggers to be disabled
        retval = self._wait_devices_propty(
            self._trigdevs, 'State-Sts', _TIConst.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO Mags Triggers to be disabled timed '\
                   'out without success! Verify BO Mags Triggers!'
            return [False, text, retval[1]]

        # wait duration of a ramp for PS change opmode
        _time.sleep(0.5)

        # set slowref
        self._set_devices_propty(
            self._psdevs, 'OpMode-Sel', _PSConst.OpMode.SlowRef)

        # wait for PS change opmode
        retval = self._wait_devices_propty(
            self._psdevs, 'OpMode-Sts', _PSConst.States.SlowRef,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO PS to be in OpMode SlowRef '\
                   'timed out without success! Verify BO PS!'
            return [False, text, retval[1]]

        # set current to zero
        self._set_devices_propty(self._psdevs, 'Current-SP', 0.0)

        # wait current change to zero
        retval = self._wait_devices_propty(
            self._psdevs, 'Current-RB', 0.0,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO PS to be with current zero '\
                   'timed out without success! Verify BO PS!'
            return [False, text, retval[1]]

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        # set rmpwfm
        self._set_devices_propty(
            self._psdevs, 'OpMode-Sel', _PSConst.OpMode.RmpWfm)

        # wait for PS change opmode
        retval = self._wait_devices_propty(
            self._psdevs, 'OpMode-Sts', _PSConst.States.RmpWfm,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO PS to be in OpMode RmpWfm '\
                   'timed out without success! Verify BO PS!'
            return [False, text, retval[1]]

        # set WfmUpdateAuto
        self._set_devices_propty(
            self._psdevs, 'WfmUpdateAuto-Sel', _PSConst.DsblEnbl.Enbl)

        # wait for PS WfmUpdateAuto to be enabled
        retval = self._wait_devices_propty(
            self._psdevs, 'WfmUpdateAuto-Sts', _PSConst.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO PS WfmUpdateAuto to be enable '\
                   'timed out without success! Verify BO PS!'
            return [False, text, retval[1]]

        # configure trigger source
        values = [dev.source_options.index('RmpBO') for dev in self._trigdevs]
        self._set_devices_propty(self._trigdevs, 'Src-Sel', values)

        # wait for trigger source to be configured
        retval = self._wait_devices_propty(
            self._trigdevs, 'Src-Sts', values, timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO Mags Triggers to be in RmpBO event '\
                   'timed out without success! Verify BO Mags Triggers!'
            return [False, text, retval[1]]

        # enable triggers
        self._set_devices_propty(
            self._trigdevs, 'State-Sel', _TIConst.DsblEnbl.Enbl)

        # wait for triggers to be enable
        retval = self._wait_devices_propty(
            self._trigdevs, 'State-Sts', _TIConst.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO Mags Triggers to be enable timed '\
                   'out without success! Verify BO Mags Triggers!'
            return [False, text, retval[1]]

        return True, '', []


class BORFRampStandbyHandler(_BaseHandler):
    """BO RF Ramp standby mode handler for injection procedure."""

    def __init__(self):
        self.llrf = ASLLRF(ASLLRF.DEVICES.BO)
        self.rmptrg = Trigger('BO-Glob:TI-LLRF-Rmp')

        devices = (self.llrf, self.rmptrg)

        self._on_values = dict()
        self._on_values[self.llrf] = {
            'RmpEnbl-Sts': _Const.DsblEnbl.Enbl,
            'RmpReady-Mon': _Const.DsblEnbl.Enbl}
        self._on_values[self.rmptrg] = {
            'State-Sts': _Const.DsblEnbl.Enbl}

        # call base class constructor
        super().__init__('', devices)

    def cmd_turn_off(self):
        """Turn off."""
        # set RF ramp to disable
        self.llrf.rmp_enable = _Const.DsblEnbl.Dsbl

        # wait for RF ramp to be disable
        retval = self._wait_devices_propty(
            self.llrf, 'RmpEnbl-Sts', _Const.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO RF Ramp to be disabled timed '\
                   'out without success! Verify BO RF Ramp!'
            return [False, text, retval[1]]

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        # set RF ramp to enabled
        self.llrf.rmp_enable = _Const.DsblEnbl.Enbl

        # wait for RF ramp to be enabled
        retval = self._wait_devices_propty(
            self.llrf, 'RmpEnbl-Sts', _Const.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for BO RF Ramp to be enabled timed '\
                   'out without success! Verify BO RF Ramp!'
            return [False, text, retval[1]]

        return True, '', []


class InjBOStandbyHandler(_BaseHandler):
    """InjBO Event standby mode handler for injection procedure."""

    def __init__(self):
        self.evg = EVG()
        self.injboevt = Event('InjBO')

        devices = (self.evg, self.injboevt)

        self._on_values = {
            self.injboevt: {'Mode-Sts': Event.MODES.index('Continuous')}}

        # call base class constructor
        super().__init__('', devices)

    def cmd_turn_off(self):
        """Turn off."""
        # disable injbo
        self.injboevt.mode = 'Disable'

        # wait for injbo to be disabled
        retval = self._wait_devices_propty(
            self.injboevt, 'Mode-Sts', Event.MODES.index('Disable'),
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for InjBO Event to be disabled timed '\
                   'out without success! Verify InjBO Event!'
            return [False, text, retval[1]]

        # update events
        _time.sleep(1)
        self.evg.cmd_update_events()
        _time.sleep(1)

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        # set injbo to Continuous table
        self.injboevt.mode = 'Continuous'

        # wait for injbo to be in Continuous Table
        retval = self._wait_devices_propty(
            self.injboevt, 'Mode-Sts', Event.MODES.index('Continuous'),
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for InjBO Event to be in Continuous table '\
                   'timed out without success! Verify InjBO Event!'
            return [False, text, retval[1]]

        # update events
        self.evg.cmd_update_events()

        return True, '', []


class LILLRFStandbyHandler(_BaseHandler):
    """LI LLRF standby mode handler for injection procedure."""

    WAIT_2_TURNON = 2  # [s]

    def __init__(self):
        """Init."""

        # create devices
        devices = list()
        for dev in DevLILLRF.DEVICES.ALL:
            devices.append(DevLILLRF(dev))
        devices = tuple(devices)

        self._on_values = dict()
        for dev in devices:
            self._on_values[dev] = {
                'GET_INTEGRAL_ENABLE': _Const.DsblEnbl.Enbl,
                'GET_FB_MODE': _Const.DsblEnbl.Enbl}

        # call base class constructor
        super().__init__('', devices)

    def cmd_turn_off(self):
        """Turn off."""
        # turn feedback off
        self._set_devices_propty(
            self.devices, 'SET_FB_MODE', _Const.DsblEnbl.Dsbl)

        # wait for feedback to turn off
        retval = self._wait_devices_propty(
            self.devices, 'GET_FB_MODE', _Const.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI LLRF Feedback Mode to be off '\
                   'timed out without success! Verify LI LLRF!'
            return [False, text, retval[1]]

        # turn integral off
        self._set_devices_propty(
            self.devices, 'SET_INTEGRAL_ENABLE', _Const.DsblEnbl.Dsbl)

        # wait for integral to turn off
        retval = self._wait_devices_propty(
            self.devices, 'GET_INTEGRAL_ENABLE', _Const.DsblEnbl.Dsbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI LLRF Integral Mode to be off '\
                   'timed out without success! Verify LI LLRF!'
            return [False, text, retval[1]]

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        # wait for some InjBO pulses
        _time.sleep(LILLRFStandbyHandler.WAIT_2_TURNON)

        # turn integral on
        self._set_devices_propty(
            self.devices, 'SET_INTEGRAL_ENABLE', _Const.DsblEnbl.Enbl)

        # wait for integral to turn on
        retval = self._wait_devices_propty(
            self.devices, 'GET_INTEGRAL_ENABLE', _Const.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI LLRF Integral Mode to be on '\
                   'timed out without success! Verify LI LLRF!'
            return [False, text, retval[1]]

        # turn feedback on
        self._set_devices_propty(
            self.devices, 'SET_FB_MODE', _Const.DsblEnbl.Enbl)

        # wait for feedback to turn on
        retval = self._wait_devices_propty(
            self.devices, 'GET_FB_MODE', _Const.DsblEnbl.Enbl,
            timeout=3, return_prob=True)
        if not retval[0]:
            text = 'Check for LI LLRF Feedback Mode to be on '\
                   'timed out without success! Verify LI LLRF!'
            return [False, text, retval[1]]

        return True, '', []


class InjSysStandbyHandler(_Devices):
    """Injection system standy mode handler."""

    DEF_ON_ORDER = _InjConst.INJSYS_DEF_ON_ORDER
    DEF_OFF_ORDER = _InjConst.INJSYS_DEF_OFF_ORDER
    HANDLER_DESC = {
        'as_pu': 'AS PU (Septa, Kickers and Modulators)',
        'bo_ps': 'BO PS Ramp',
        'bo_rf': 'BO RF Ramp',
        'injbo': 'TI InjBO Event',
        'li_rf': 'LI LLRF (Klystrons Loop)',
    }

    def __init__(self):
        """Init."""
        devs = {
            'as_pu': ASPUStandbyHandler(),
            'bo_ps': BOPSRampStandbyHandler(),
            'bo_rf': BORFRampStandbyHandler(),
            'injbo': InjBOStandbyHandler(),
            'li_rf': LILLRFStandbyHandler(),
        }
        self._dev_refs = devs
        self._on_order = InjSysStandbyHandler.DEF_ON_ORDER
        self._off_order = InjSysStandbyHandler.DEF_OFF_ORDER
        self._is_running = ''
        self._done = list()
        self._result = None
        self._thread = None
        self._abort = False

        self._on_values = dict()
        for dev in devs.values():
            self._on_values.update(dev.on_values)

        # call super init
        super().__init__('', tuple(devs.values()))

    @property
    def on_order(self):
        """On setpoints order."""
        return _dcopy(self._on_order)

    @on_order.setter
    def on_order(self, value):
        self._on_order = value

    @property
    def off_order(self):
        """Off setpoints order."""
        return _dcopy(self._off_order)

    @off_order.setter
    def off_order(self, value):
        self._off_order = value

    @property
    def is_on(self):
        """Is on."""
        for dev in self.devices:
            if not dev.is_on:
                return False
        return True

    @property
    def on_values(self):
        """Dict with PVs to be monitored and their respective 'on' value."""
        return self._on_values

    @property
    def is_running(self):
        """Is running."""
        return self._is_running

    @property
    def done(self):
        """Return commands already done."""
        return self._done

    @property
    def result(self):
        """Command result."""
        return self._result

    def cmd_abort(self):
        """Abort command."""
        if self._is_running:
            self._abort = True
        return True

    def cmd_reset_comm_order(self):
        """Reset commands order."""
        self.on_order = InjSysStandbyHandler.DEF_ON_ORDER
        self.off_order = InjSysStandbyHandler.DEF_OFF_ORDER
        return True

    def cmd_turn_on(self, run_in_thread=False):
        """Turn on."""
        return self._command_base('on', run_in_thread)

    def cmd_turn_off(self, run_in_thread=False):
        """Turn off."""
        return self._command_base('off', run_in_thread)

    def get_dev_state(self, devnames):
        """
        Return the state, on (True) or off (False), for each device in
        devnames.

        Parameters
        ----------
        devnames: list [str]
            A list of strings with the names of the handler devices

        Returns
        -------
        states: list [bool]
            A list of booleans with the state of each handler devices
        """
        states = list()
        for devname in devnames:
            dev = self._dev_refs[devname]
            states.append(dev.is_on)
        return states

    # --- private methods ---

    def _command_base(self, cmmtype, run_in_thread):
        if self.is_running:
            _log.error('Commands already running!')
            return

        if run_in_thread:
            self._is_running = cmmtype
            self._thread = _Thread(
                target=self._command_target, args=[cmmtype, ], daemon=True)
            self._thread.start()
            return

        self._command_target(cmmtype)
        return self._result

    def _command_target(self, cmdtype):
        order = getattr(self, cmdtype+'_order')

        self._result = None
        self._done = list()
        retval = None
        for handler_name in order:
            if self._abort:
                self._abort = False
                break
            handler = self._dev_refs[handler_name]
            func = getattr(handler, 'cmd_turn_'+cmdtype)
            retval = func()
            if retval[0]:
                self._done.append(handler_name)
            else:
                break

        self._result = retval

        self._is_running = ''


class InjSysPUModeHandler(_Devices, _Callback):
    """Device to control pulsed magnets configuration for injection."""

    _DEF_TIMEOUT = 10  # [s]
    _DEF_SLEEP = 0.1  # [s]
    SI_DPKCKR_DEFKICK = -6.7  # [mrad]
    TS_POSANG_DEFDELTA = 2.5  # [mrad]
    SI_DPKCKR_DLYREF = 36800000  # [count]

    def __init__(self, print_log=True, callback=None):
        """Init."""
        self.pudpk = PowerSupplyPU(PowerSupplyPU.DEVICES.SI_INJ_DPKCKR)
        self.punlk = PowerSupplyPU(PowerSupplyPU.DEVICES.SI_INJ_NLKCKR)
        self.trigdpk = Trigger('SI-01SA:TI-InjDpKckr')
        self.trignlk = Trigger('SI-01SA:TI-InjNLKckr')
        self.posang = PosAng(PosAng.DEVICES.TS)
        self.delta_posang = self.TS_POSANG_DEFDELTA
        self.dpkckr_dlyref = self.SI_DPKCKR_DLYREF
        self.dpkckr_kick = self.SI_DPKCKR_DEFKICK
        devices = (
            self.pudpk, self.punlk,
            self.trigdpk, self.trignlk,
            self.posang,
        )
        self._print_log = print_log
        self._abort = _Flag()

        # call super init
        _Devices.__init__(self, '', devices)
        _Callback.__init__(self, callback=callback)

    @property
    def is_trigdpk_onaxis(self):
        """Whether DpK trigger delay raw is above dpkckr_dlyref."""
        return self.trigdpk.delay_raw > self.dpkckr_dlyref

    @property
    def is_accum(self):
        """If configuration is Accumulation."""
        if not self.connected:
            return False
        dpk_off = (self.pudpk.pwrstate == PowerSupplyPU.PWRSTATE.Off) or \
            (self.pudpk.pulse == PowerSupplyPU.PULSTATE.Off)
        nlk_on = (self.punlk.pwrstate == PowerSupplyPU.PWRSTATE.On) and \
            (self.punlk.pulse == PowerSupplyPU.PULSTATE.On)
        return dpk_off & nlk_on

    def cmd_switch_to_accum(self):
        """Switch to Accumulation."""
        self._abort.clear()
        self._update_status('Switching to PU Accumulation config...')
        if not self._check_pu_interlocks():
            return False
        if self._check_abort():
            return False

        # if previously in on-axis, do delta angle x
        if self.is_onaxis:
            if not self._do_delta_posang(-self.delta_posang):
                return False

        # set pulsed magnet pwrstate and pulse
        proced = (
            (self.pudpk.cmd_turn_off_pulse, 'turn DpK pulse off.'),
            (self.pudpk.cmd_turn_off, 'turn DpK off.'),
            (self.punlk.cmd_turn_on, 'turn NLK on.'),
            (self.punlk.cmd_turn_on_pulse, 'turn NLK pulse on.'),
        )
        if not self._do_procedures(proced):
            return False

        self._update_status('PU mode switched to Accumulation!')
        return True

    @property
    def is_optim(self):
        """If configuration is Optimization."""
        if not self.connected:
            return False
        dpk_evt_ok = self.trigdpk.source == self.trignlk.source
        dpk_dly_ok = not self.is_trigdpk_onaxis
        dpk_kck_ok = abs(self.pudpk.strength - self.dpkckr_kick) < 1e-3
        dpk_on = (self.pudpk.pwrstate == PowerSupplyPU.PWRSTATE.On) and \
            (self.pudpk.pulse == PowerSupplyPU.PULSTATE.On)
        nlk_on = (self.punlk.pwrstate == PowerSupplyPU.PWRSTATE.On) and \
            (self.punlk.pulse == PowerSupplyPU.PULSTATE.On)
        return dpk_evt_ok & dpk_dly_ok & dpk_kck_ok & dpk_on & nlk_on

    def cmd_switch_to_optim(self):
        """Switch to Optimization."""
        self._abort.clear()
        self._update_status('Switching to PU Optimization config...')
        if not self._check_pu_interlocks():
            return False
        if self._check_abort():
            return False

        # if previously in on-axis, do delta angle x
        if self.is_onaxis:
            if not self._do_delta_posang(-self.delta_posang):
                return False

        if self._check_abort():
            return False

        # configure DpK trigger
        if self.is_trigdpk_onaxis:
            delay = self.trigdpk.delay_raw - self.dpkckr_dlyref
            if not self._config_dpk_trigger(delayraw=delay):
                return False

        # set DpK Kick
        if not self._config_dpk_kick():
            return False

        # set pulsed magnet pwrstate and pulse
        proced = (
            (self.pudpk.cmd_turn_on, 'turn DpK on.'),
            (self.punlk.cmd_turn_on, 'turn NLK on.'),
            (self.pudpk.cmd_turn_on_pulse, 'turn DpK pulse on.'),
            (self.punlk.cmd_turn_on_pulse, 'turn NLK pulse on.'),
        )
        if not self._do_procedures(proced):
            return False

        self._update_status('PU mode switched to Optimization!')
        return True

    @property
    def is_onaxis(self):
        """If configuration is OnAxis."""
        if not self.connected:
            return False
        dpk_evt_ok = self.trigdpk.source == self.trignlk.source
        dpk_dly_ok = self.is_trigdpk_onaxis
        dpk_kck_ok = abs(self.pudpk.strength - self.dpkckr_kick) < 1e-3
        dpk_on = (self.pudpk.pwrstate == PowerSupplyPU.PWRSTATE.On) and \
            (self.pudpk.pulse == PowerSupplyPU.PULSTATE.On)
        nlk_off = (self.punlk.pwrstate == PowerSupplyPU.PWRSTATE.Off) or \
            (self.punlk.pulse == PowerSupplyPU.PULSTATE.Off)
        return dpk_evt_ok & dpk_dly_ok & dpk_kck_ok & dpk_on & nlk_off

    def cmd_switch_to_onaxis(self):
        """Switch to OnAxis."""
        self._abort.clear()
        self._update_status('Switching to PU OnAxis config...')
        if not self._check_pu_interlocks():
            return False
        if self._check_abort():
            return False

        # if not previously in on-axis, do delta angle x
        if not self.is_onaxis:
            if not self._do_delta_posang(self.delta_posang):
                return False

        if self._check_abort():
            return False

        # configure DpK trigger
        if not self.is_trigdpk_onaxis:
            delay = self.trigdpk.delay_raw + self.dpkckr_dlyref
            if not self._config_dpk_trigger(delayraw=delay):
                return False

        # set DpK Kick
        if not self._config_dpk_kick():
            return False

        # set pulsed magnet pwrstate and pulse
        proced = (
            (self.punlk.cmd_turn_off_pulse, 'turn NLK pulse off.'),
            (self.punlk.cmd_turn_off, 'turn NLK off.'),
            (self.pudpk.cmd_turn_on, 'turn DpK on.'),
            (self.pudpk.cmd_turn_on_pulse, 'turn DpK pulse on.'),
        )
        if not self._do_procedures(proced):
            return False

        self._update_status('PU mode switched to OnAxis!')
        return True

    def _check_pu_interlocks(self):
        if not self.pudpk.interlock_ok:
            self._update_status('ERR: DpK interlocks not ok. Aborted.')
            return False
        if not self.punlk.interlock_ok:
            self._update_status('ERR: NLK interlocks not ok. Aborted.')
            return False
        return True

    def _do_delta_posang(self, delta):
        if self.posang.need_ref_update:
            self.posang.cmd_update_reference()
        _time.sleep(InjSysPUModeHandler._DEF_SLEEP)
        desval = self.posang.delta_angx + delta
        self.posang.delta_angx += delta
        if not self._wait(self.posang, 'delta_angx', desval):
            self._update_status('ERR:Could not do delta AngX.')
            return False
        return True

    def _config_dpk_trigger(self, delayraw):
        # set DpK trigger event to NLK trigger event
        dpk_event = self.trignlk.source
        self.trigdpk.source = dpk_event
        if not self._wait(self.trigdpk, 'source', dpk_event):
            self._update_status('ERR:Could not set DpK trigger event.')
            return False
        # set DpK trigger DelayRaw
        self.trigdpk.delay_raw = delayraw
        if not self._wait(self.trigdpk, 'delay_raw', delayraw):
            self._update_status('ERR:Could not set DpK trigger delay.')
            return False
        return True

    def _config_dpk_kick(self):
        self.pudpk.strength = self.dpkckr_kick
        if not self._wait(self.pudpk, 'strength', self.dpkckr_kick):
            self._update_status(
                'ERR:Could not set DpK Kick to '
                f'{self.dpkckr_kick:.1f}mrad.')
            return False
        return True

    def _do_procedures(self, procedures):
        for fun, msg in procedures:
            if self._check_abort():
                return False
            if not fun(timeout=2):
                self._update_status('ERR:Could not ' + msg)
                return False
            _time.sleep(0.5)
        return True

    # -------- thread help methods --------

    def cmd_abort(self):
        """Abort PU mode change."""
        self._abort.set()
        self._update_status('WARN:Abort received for PU Mode change.')
        return True

    def _check_abort(self):
        """Clear abort flag."""
        if not self._abort.is_set():
            return False
        self._abort.clear()
        self._update_status('ERR:Aborted PU Mode change.')
        return True

    # ---------- check sp -----------

    def _wait(
            self, device, prop, desired, tolerance=1e-3, timeout=_DEF_TIMEOUT):
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if abs(getattr(device, prop) - desired) < tolerance:
                return True
            _time.sleep(InjSysPUModeHandler._DEF_SLEEP)
        return False

    # ---------- logging -----------

    def _update_status(self, status):
        if self._print_log:
            print(status)
        self.run_callbacks(status)



#!/usr/bin/env python-sirius
"""Magnet class test module."""

from unittest import TestCase
import numpy

from siriuspy import util
from siriuspy.pwrsupply.csdev import DEF_WFMSIZE_OTHERS as \
    _DEF_WFMSIZE_OTHERS
from siriuspy.ramp import waveform
from siriuspy.ramp.waveform import WaveformParam
from siriuspy.ramp.waveform import WaveformDipole
from siriuspy.ramp.waveform import Waveform


PUB_INTERFACE = (
    'WaveformParam',
    'WaveformDipole',
    'Waveform',
)


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(waveform, PUB_INTERFACE)
        self.assertTrue(valid)


class TestWaveformParam(TestCase):
    """Test WaveformParam class."""

    PUB_INTERFACE = (
        'N',
        'duration',
        'start_value',
        'rampup1_start_time',
        'rampup1_start_value',
        'rampup2_start_time',
        'rampup2_start_value',
        'rampdown_start_time',
        'rampdown_start_value',
        'rampdown_stop_value',
        'rampdown_stop_time',
        'rampup_smooth_intvl',
        'rampup_smooth_value',
        'rampdown_smooth_intvl',
        'rampdown_smooth_value',
        'rampup1_slope',
        'rampup2_slope',
        'rampdown_slope',
        'eval_at',
    )

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            WaveformParam, TestWaveformParam.PUB_INTERFACE)
        self.assertTrue(valid)

    def test_rampup1_start_time(self):
        """Test rampup1_start_time."""
        wfm = WaveformParam()
        val1 = 0.0
        val2 = wfm.rampup1_start_time
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_rampup2_start_time(self):
        """Test rampup2_start_time."""
        wfm = WaveformParam()
        val1 = wfm.rampup1_start_time
        val2 = wfm.rampup2_start_time
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_rampdown_start_time(self):
        """Test rampdown_start_time."""
        wfm = WaveformParam()
        val1 = wfm.rampup2_start_time
        val2 = wfm.rampdown_start_time
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_rampdown_stop_time(self):
        """Test rampdown_stop_time."""
        wfm = WaveformParam()
        val1 = wfm.rampdown_start_time
        val2 = wfm.rampdown_stop_time
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_duration(self):
        """Test duration."""
        wfm = WaveformParam()
        val1 = wfm.rampdown_stop_time
        val2 = wfm.duration
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_start_value(self):
        """Test start_value."""
        wfm = WaveformParam()
        val1 = 0.0
        val2 = wfm.start_value
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 >= val1)

    def test_rampup1_start_value(self):
        """Test rampup1_start_value."""
        wfm = WaveformParam()
        val1 = wfm.start_value
        val2 = wfm.rampup1_start_value
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_rampup2_start_value(self):
        """Test rampup2_start_value."""
        wfm = WaveformParam()
        val1 = wfm.rampup1_start_value
        val2 = wfm.rampup2_start_value
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_rampdown_start_value(self):
        """Test rampdown_start_value."""
        wfm = WaveformParam()
        val1 = wfm.rampup2_start_value
        val2 = wfm.rampdown_start_value
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 > val1)

    def test_rampdown_stop_value(self):
        """Test rampdown_stop_value."""
        wfm = WaveformParam()
        val1 = wfm.rampdown_start_value
        val2 = wfm.rampdown_stop_value
        self.assertIsInstance(val2, float)
        self.assertTrue(val2 < val1)

    def test_rampup_smooth_intvl(self):
        """Test rampup_smooth_intvl."""
        wfm = WaveformParam()
        val1 = wfm.rampup_smooth_intvl
        self.assertIsInstance(val1, float)
        self.assertTrue(val1 > 0)

    def test_rampup_smooth_value(self):
        """Test rampup_smooth_value."""
        wfm = WaveformParam()
        val1 = wfm.rampup_smooth_value
        self.assertIsInstance(val1, float)

    def test_rampdown_smooth_intvl(self):
        """Test rampdown_smooth_intvl."""
        wfm = WaveformParam()
        val1 = wfm.rampdown_smooth_intvl
        self.assertIsInstance(val1, float)
        self.assertTrue(val1 > 0)

    def test_rampdown_smooth_value(self):
        """Test rampdown_smooth_value."""
        wfm = WaveformParam()
        val1 = wfm.rampdown_smooth_value
        self.assertIsInstance(val1, float)


class TestWaveformDipole(TestCase):
    """Test WaveformDipole class."""

    PUB_INTERFACE = (
        'waveform',
        'start_energy',
        'rampup1_start_energy',
        'rampup2_start_energy',
        'rampdown_start_energy',
        'rampdown_stop_energy',
        'rampup_smooth_energy',
        'rampdown_smooth_energy',
        'start_value',
        'rampup1_start_value',
        'rampup2_start_value',
        'rampdown_start_value',
        'rampdown_stop_value',
        'rampup_smooth_value',
        'rampdown_smooth_value',
    )

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            WaveformDipole, TestWaveformDipole.PUB_INTERFACE)
        self.assertTrue(valid)

    def _test_constructor(self):
        """Test class constructor."""
        # TODO: implement using mock to substitute server
        # default arguments
        wfm = WaveformDipole()
        curr = wfm.currents
        self.assertIsInstance(curr, numpy.ndarray)
        self.assertEqual(len(curr), _DEF_WFMSIZE_OTHERS)
        stren = wfm.strengths
        self.assertIsInstance(stren, list)
        self.assertEqual(len(stren), _DEF_WFMSIZE_OTHERS)

    def test_waveform(self):
        """Test waveform."""
        # TODO: implement it!

    def test_update(self):
        """Test update."""
        # TODO: implement it!


class TestWaveform(TestCase):
    """Test Waveform class."""

    PUB_INTERFACE = (
        'duration',
        'update',
    )

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            Waveform, TestWaveform.PUB_INTERFACE)
        self.assertTrue(valid)

    def test_duration(self):
        """Test duration."""
        # TODO: implement it!

    def _test_update(self):
        """Test update."""
        # TODO: implement using mock to substitute server
        bo_b = WaveformDipole()
        defoc = Waveform(psname='BO-Fam:MA-QD', dipole=bo_b)
        curr1 = defoc.currents
        bo_b.rampup1_start_energy *= 1.01
        curr2 = defoc.currents
        self.assertTrue((curr1 != curr2).any())



"""UDC class."""

import time as _time
import numpy as _np

from . import prucparms as _prucparms

from ..bsmp.constants import UDC_MAX_NR_DEV as _UDC_MAX_NR_DEV
from ..bsmp.factory import PSBSMPFactory as _PSBSMPFactory


class UDC:
    """UDC."""

    _prucparms = {
        'FBP': _prucparms.PRUCParmsFBP,
        'FBP_DCLink': _prucparms.PRUCParmsFBP_DCLink,
        'FAC_2S_DCDC': _prucparms.PRUCParmsFAC_2S_DCDC,
        'FAC_2P4S_DCDC': _prucparms.PRUCParmsFAC_2P4S_DCDC,
        'FAC_DCDC': _prucparms.PRUCParmsFAC_DCDC,
        'FAC_2S_ACDC': _prucparms.PRUCParmsFAC_2S_ACDC,
        'FAC_2P4S_ACDC': _prucparms.PRUCParmsFAC_2P4S_ACDC,
        'FAP': _prucparms.PRUCParmsFAP,
        'FAP_4P': _prucparms.PRUCParmsFAP_4P,
        'FAP_2P2S': _prucparms.PRUCParmsFAP_2P2S,
    }

    _soft_def = _np.zeros(_UDC_MAX_NR_DEV)
    _soft_def_int = _np.zeros(_UDC_MAX_NR_DEV, dtype=int)

    def __init__(self, pru, psmodel, device_ids):
        """Init."""
        self._pru = pru
        self._devs_ids = device_ids
        self._psmodel = psmodel
        self._devs_bsmp = self._create_bsmp_connectors()
        self._dev_first, self._dev_second = \
            self._get_devs_first_second()
        self._is_fbp = self._psmodel == 'FBP'

    def _get_devs_first_second(self):
        ids = sorted(self._devs_ids)
        dev_first = self._devs_bsmp[ids[0]]
        dev_second = self._devs_bsmp[ids[4]] if len(ids) > 4 else None
        return dev_first, dev_second

    @property
    def pru(self):
        """Return PRU used for communication with UDC."""
        return self._pru

    @property
    def psmodel(self):
        """Return power supply model associated with UDC."""
        return self._psmodel

    @property
    def device_ids(self):
        """."""
        return self._devs_ids

    @property
    def prucparms(self):
        """Return PRU Controller parameters object."""
        prucparms_class = UDC._prucparms[self._psmodel]
        return prucparms_class()

    @property
    def CONST_BSMP(self):
        """Return PSBSMP constants."""
        return self._dev_first.CONST_BSMP

    @property
    def CONST_PSBSMP(self):
        """Return PSBSMP constants."""
        return self.prucparms.CONST_PSBSMP

    def reset(self, **kwargs):
        """Reset UDC."""
        # turn off all power supplies (NOTE: or F_RESET_UDC does not work)
        for dev in self._devs_bsmp.values():
            ack, data = dev.execute_function(
                func_id=self.CONST_PSBSMP.F_TURN_OFF)
            if ack != self.CONST_BSMP.ACK_OK:
                dev_id = dev.channel.address
                print(('UDC.reset: could not turn off '
                       'device id:{} !').format(dev_id))
                return ack, data

        # reset UDC proper.
        self._dev_first.execute_function(
            func_id=self.CONST_PSBSMP.F_RESET_UDC, **kwargs, read_flag=False)

        return ack, data

    def bufsample_disable(self):
        """Disable DSP from writting to bufsample curve."""
        for dev in self._devs_bsmp.values():
            ack, data = dev.scope_disable()
            if ack != self.CONST_BSMP.ACK_OK:
                return ack, data
        # a single sleep for all devices
        sleep_time = self._dev_first._sleep_disable_bufsample = 0.5  # [s]
        _time.sleep(sleep_time)
        return ack, data

    def bufsample_enable(self):
        """Enable DSP from writting to bufsample curve."""
        for dev in self._devs_bsmp.values():
            ack, data = dev.scope_enable()
            if ack != self.CONST_BSMP.ACK_OK:
                return ack, data
        return ack, data

    def parse_firmware_version(self, version):
        """Process firmware version from BSMP device."""
        # uses first BSMP device
        dev = self._devs_bsmp[next(iter(self._devs_bsmp))]  # first dev
        return dev.parse_firmware_version(version)

    # NOTE: this unused function can be deleted in the future
    # def reset_groups_of_variables(self, groups):
    #     """Reset groups of variables."""
    #     for dev in self._devs_bsmp.values():
    #         dev.reset_groups_of_variables(groups=groups)

    def add_groups_of_variables(self):
        """Add psmodel-specific variables groups in entities (no comm)."""
        # build varids for groups with G_IDS >= 3
        gids = sorted(tuple(self.prucparms.groups.keys()))
        groups = [self.prucparms.groups[gid] for gid in gids[3:]]

        for dev in self._devs_bsmp.values():
            # reset group list, adding only default groups 0, 1 and 2.
            dev.entities.reset_group()
            # add psmodel-specific groups
            for var_ids in groups:
                dev.entities.add_group(var_ids)

    def __getitem__(self, index):
        """Return BSMP."""
        return self._devs_bsmp[index]

    def __len__(self):
        """."""
        return len(self._devs_bsmp)

    # --- SOFBCurrent methods

    def sofb_current_rb_get(self):
        """Return SOFBCurrent-RB."""
        if self._is_fbp:
            dev1, dev2 = self._dev_first, self._dev_second
            val1 = dev1.sofb_ps_setpoint
            val2 = dev2.sofb_ps_setpoint if dev2 else UDC._soft_def
            if val1 is not None and val2 is not None:
                return _np.concatenate((val1, val2))
        return None

    def sofb_current_refmon_get(self):
        """Return SOFBCurrentRef-Mon."""
        if self._is_fbp:
            dev1, dev2 = self._dev_first, self._dev_second
            val1 = dev1.sofb_ps_reference
            val2 = dev2.sofb_ps_reference if dev2 else UDC._soft_def
            if val1 is not None and val2 is not None:
                return _np.concatenate((val1, val2))
        return None

    def sofb_current_mon_get(self):
        """Return SOFBCurrent-Mon."""
        if self._is_fbp:
            dev1, dev2 = self._dev_first, self._dev_second
            val1 = dev1.sofb_ps_iload
            val2 = dev2.sofb_ps_iload if dev2 else UDC._soft_def
            if val1 is not None and val2 is not None:
                return _np.concatenate((val1, val2))
        return None

    def sofb_current_readback_ref_get(self):
        """Return SOFBCurrentRef-Mon, readback of last setpoint."""
        if self._is_fbp:
            dev1, dev2 = self._dev_first, self._dev_second
            val1 = dev1.sofb_ps_readback_ref
            val2 = dev2.sofb_ps_readback_ref if dev2 else UDC._soft_def
            if val1 is not None and val2 is not None:
                return _np.concatenate((val1, val2))
        return None

    def sofb_func_return_get(self):
        """Return function return from last SOFB communication."""
        if self._is_fbp:
            dev1, dev2 = self._dev_first, self._dev_second
            val1 = dev1.sofb_ps_func_return
            val2 = dev2.sofb_ps_func_return if dev2 else UDC._soft_def_int
            return _np.concatenate((val1, val2))
        return None

    def sofb_current_set(self, value):
        """Set SOFB Current."""
        if self._is_fbp:
            # set value
            self._dev_first.sofb_ps_setpoint_set(value[:_UDC_MAX_NR_DEV])
            if self._dev_second:
                self._dev_second.sofb_ps_setpoint_set(value[_UDC_MAX_NR_DEV:])

    def sofb_update(self):
        """Update sofb."""
        if self._is_fbp:
            self._dev_first.sofb_update()
            if self._dev_second:
                self._dev_second.sofb_update()

    # --- private methods

    def _create_bsmp_connectors(self):
        devices = dict()
        for dev_id in self._devs_ids:
            devices[dev_id] = _PSBSMPFactory.create(
                self._psmodel, dev_id, self._pru)
        return devices



"""Class of the Response Matrix."""

import os as _os
from copy import deepcopy as _dcopy
import logging as _log
from functools import partial as _part
from threading import Thread as _Thread

import numpy as _np

from .base_class import BaseClass as _BaseClass


class BaseMatrix(_BaseClass):
    """."""


class EpicsMatrix(BaseMatrix):
    """Class of the Response Matrix."""

    def __init__(self, acc, prefix='', callback=None):
        """Initialize the instance."""
        super().__init__(acc, prefix=prefix, callback=callback)
        self._sofb = None
        self.select_items = {
            'bpmx': _np.ones(self._csorb.nr_bpms, dtype=bool),
            'bpmy': _np.ones(self._csorb.nr_bpms, dtype=bool),
            'ch': _np.ones(self._csorb.nr_ch, dtype=bool),
            'cv': _np.ones(self._csorb.nr_cv, dtype=bool),
            }
        self.selection_pv_names = {
            'ch': 'CHEnblList-RB',
            'cv': 'CVEnblList-RB',
            'bpmx': 'BPMXEnblList-RB',
            'bpmy': 'BPMYEnblList-RB',
            }
        if self.isring:
            self.select_items['rf'] = _np.zeros(1, dtype=bool)
            self.selection_pv_names['rf'] = 'RFEnbl-Sts'
        self._respmat_mode = self._csorb.RespMatMode.Full
        self.min_sing_val = self._csorb.MIN_SING_VAL
        self.tikhonov_reg_const = self._csorb.TIKHONOV_REG_CONST
        self.respmat = _np.zeros(
            [2*self._csorb.nr_bpms, self._csorb.nr_corrs], dtype=float)
        self.inv_respmat = self.respmat.copy().T
        self.respmat_processed = self.respmat.copy().T
        self._load_respmat()

    @property
    def sofb(self):
        """."""
        return self._sofb

    @sofb.setter
    def sofb(self, sofb):
        self._sofb = sofb

    @property
    def bpm_enbllist(self):
        """."""
        sel_ = self.select_items
        return _np.hstack([sel_['bpmx'], sel_['bpmy']])

    @property
    def corrs_enbllist(self):
        """."""
        sel_ = self.select_items
        seleccor = _np.hstack([sel_['ch'], sel_['cv']])
        if self.isring:
            seleccor = _np.hstack([seleccor, sel_['rf']])
        return seleccor

    def get_map2write(self):
        """Get the write methods of the class."""
        dbase = {
            'RespMat-SP': self.set_respmat,
            'RespMatMode-Sel': self.set_respmat_mode,
            'CHEnblList-SP': _part(self.set_enbllist, 'ch'),
            'CVEnblList-SP': _part(self.set_enbllist, 'cv'),
            'BPMXEnblList-SP': _part(self.set_enbllist, 'bpmx'),
            'BPMYEnblList-SP': _part(self.set_enbllist, 'bpmy'),
            'MinSingValue-SP': self.set_min_sing_value,
            'TikhonovRegConst-SP': self.set_tikhonov_reg_const,
            }
        if self.isring:
            dbase['RFEnbl-Sel'] = _part(self.set_enbllist, 'rf')
        return dbase

    def set_respmat_mode(self, mode):
        """Set the response matrix mode."""
        msg = 'Setting New RespMatMode.'
        self._update_log(msg)
        _log.info(msg)
        if mode not in self._csorb.RespMatMode:
            return False
        old_ = self._respmat_mode
        self._respmat_mode = mode
        if not self._calc_matrices():
            self._respmat_mode = old_
            return False
        self.run_callbacks('RespMatMode-Sts', mode)
        return True

    def set_respmat(self, mat):
        """Set the response matrix in memory and save it in file."""
        msg = 'Setting New RespMat.'
        self._update_log(msg)
        _log.info(msg)
        if mat is None:
            return False
        mat = _np.reshape(mat, [-1, self._csorb.nr_corrs])
        old_ = self.respmat.copy()
        self.respmat = mat
        if not self._calc_matrices():
            self.respmat = old_
            return False
        self._save_respmat(mat)
        self.run_callbacks('RespMat-RB', list(self.respmat.ravel()))
        return True

    def set_enbllist(self, key, val):
        """."""
        msg = 'Setting {0:s} EnblList'.format(key.upper())
        self._update_log(msg)
        _log.info(msg)

        bkup = self.select_items[key]
        new_ = _np.array(val, dtype=bool)
        if key == 'rf':
            pass
        elif new_.size >= bkup.size:
            new_ = new_[:bkup.size]
        else:
            new2_ = bkup.copy()
            new2_[:new_.size] = new_
            new_ = new2_
        self.select_items[key] = new_

        if not self._calc_matrices():
            self.select_items[key] = bkup
            return False
        if new_.size == 1:  # Deal with RF
            self.run_callbacks(self.selection_pv_names[key], bool(new_))
        else:
            self.run_callbacks(self.selection_pv_names[key], new_)
        return True

    def calc_kicks(self, orbit):
        """Calculate the kick from the orbit distortion given."""
        if orbit.size != self.inv_respmat.shape[1]:
            msg = 'ERR: Orbit and matrix size not compatible.'
            self._update_log(msg)
            _log.error(msg[5:])
            return None
        kicks = _np.dot(self.inv_respmat, orbit)
        kicks *= -1
        _Thread(
            target=self._update_dkicks, args=(kicks, ), daemon=True).start()
        return kicks

    def estimate_orbit_variation(self, kicks):
        """Estimate the orbit variation induced by kicks.

        Args:
            kicks (numpy.ndarray): Correctors kicks that will be used to
                estimate orbit distortion.

        Returns:
            dorb (numpy.ndarray): Estimated orbit.

        """
        if kicks.size != self.respmat.shape[1]:
            msg = 'ERR: kicks and matrix size not compatible.'
            self._update_log(msg)
            _log.error(msg[5:])
            return None
        return _np.dot(self.respmat, kicks)

    def _update_dkicks(self, kicks):
        kicks = kicks.copy()
        nr_ch = self._csorb.nr_ch
        nr_chcv = self._csorb.nr_chcv
        self.run_callbacks('DeltaKickCH-Mon', kicks[:nr_ch])
        self.run_callbacks('DeltaKickCV-Mon', kicks[nr_ch:nr_chcv])
        if self.isring:
            self.run_callbacks('DeltaKickRF-Mon', kicks[-1])

    def set_min_sing_value(self, num):
        """."""
        bkup = self.min_sing_val
        self.min_sing_val = float(num)
        if not self._calc_matrices():
            self.min_sing_val = bkup
            return False
        self.run_callbacks('MinSingValue-RB', self.min_sing_val)
        return True

    def set_tikhonov_reg_const(self, num):
        """."""
        bkup = self.tikhonov_reg_const
        self.tikhonov_reg_const = float(num)
        if not self._calc_matrices():
            self.tikhonov_reg_const = bkup
            return False
        self.run_callbacks('TikhonovRegConst-RB', self.tikhonov_reg_const)
        return True

    def _calc_matrices(self):
        msg = 'Calculating Inverse Matrix.'
        self._update_log(msg)
        _log.info(msg)
        selecbpm = self.bpm_enbllist
        seleccor = self.corrs_enbllist
        if not any(selecbpm):
            msg = 'ERR: No BPM selected in EnblList'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        if not any(seleccor):
            msg = 'ERR: No Corrector selected in EnblList'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        sel_mat = selecbpm[:, None] * seleccor[None, :]
        if sel_mat.size != self.respmat.size:
            return False
        mat = self.respmat.copy()
        nr_bpms = self._csorb.nr_bpms
        nr_ch = self._csorb.nr_ch
        nr_chcv = self._csorb.nr_chcv
        if self._respmat_mode != self._csorb.RespMatMode.Full:
            mat[:nr_bpms, nr_ch:nr_chcv] = 0
            mat[nr_bpms:, :nr_ch] = 0
            mat[nr_bpms:, nr_chcv:] = 0
        if self._respmat_mode == self._csorb.RespMatMode.Mxx:
            mat[nr_bpms:] = 0
        elif self._respmat_mode == self._csorb.RespMatMode.Myy:
            mat[:nr_bpms] = 0

        mat = mat[sel_mat]
        mat = _np.reshape(mat, [sum(selecbpm), sum(seleccor)])
        try:
            uuu, sing, vvv = _np.linalg.svd(mat, full_matrices=False)
        except _np.linalg.LinAlgError():
            msg = 'ERR: Could not calculate SVD'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        idcs = sing > self.min_sing_val
        singr = sing[idcs]
        nr_sv = _np.sum(idcs)
        if not nr_sv:
            msg = 'ERR: All Singular Values below minimum.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False

        # Apply Tikhonov regularization:
        regc = self.tikhonov_reg_const
        regc *= regc
        inv_s = _np.zeros(sing.size, dtype=float)
        inv_s[idcs] = singr/(singr*singr + regc)

        # calculate processed singular values
        singp = _np.zeros(sing.size, dtype=float)
        singp[idcs] = 1/inv_s[idcs]
        inv_mat = _np.dot(vvv.T*inv_s, uuu.T)
        is_nan = _np.any(_np.isnan(inv_mat))
        is_inf = _np.any(_np.isinf(inv_mat))
        if is_nan or is_inf:
            msg = 'ERR: Inverse contains nan or inf.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False

        sing_vals = _np.zeros(self._csorb.nr_svals, dtype=float)
        sing_vals[:sing.size] = sing
        self.run_callbacks('SingValuesRaw-Mon', sing_vals)
        sing_vals = _np.zeros(self._csorb.nr_svals, dtype=float)
        sing_vals[:singp.size] = singp
        self.run_callbacks('SingValues-Mon', sing_vals)
        self.run_callbacks('NrSingValues-Mon', nr_sv)
        self.inv_respmat = _np.zeros(self.respmat.shape, dtype=float).T
        self.inv_respmat[sel_mat.T] = inv_mat.ravel()
        self.run_callbacks('InvRespMat-Mon', list(self.inv_respmat.ravel()))
        self.respmat_processed = _np.zeros(self.respmat.shape, dtype=float)
        self.respmat_processed[sel_mat] = _np.dot(uuu*singp, vvv).ravel()
        self.run_callbacks('RespMat-Mon', list(self.respmat_processed.ravel()))
        msg = 'Ok!'
        self._update_log(msg)
        _log.info(msg)
        return True

    def _load_respmat(self):
        filename = self._csorb.respmat_fname
        boo = False
        if _os.path.isfile(filename):
            boo = self.set_respmat(_np.loadtxt(filename))
            if boo:
                msg = 'Loading RespMat from file.'
            else:
                msg = 'ERR: Problem loading RespMat from file.'
            self._update_log(msg)
            _log.info(msg)

    def _save_respmat(self, mat):
        path = _os.path.split(self._csorb.respmat_fname)[0]
        _os.makedirs(path, exist_ok=True)
        _np.savetxt(self._csorb.respmat_fname, mat)



#!/usr/bin/env python-sirius

"""Unittest module for search package."""

from unittest import TestCase

from siriuspy import util
from siriuspy import search

public_interface = (
    'PSSearch',
    'MASearch',
    'BPMSearch',
    'HLTimeSearch',
    'LLTimeSearch',
    'IDSearch',
    'IOCSearch',
)


class TestSearch(TestCase):
    """Test Search module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(search, public_interface)
        self.assertTrue(valid)



"""."""

from . import commands as _cmd_bsmp


class PSBSMPFactory:
    """."""

    psname_2_psbsmp = {
        'FBP': _cmd_bsmp.FBP,
        'FBP_DCLink': _cmd_bsmp.FBP_DCLink,
        'FAC_DCDC': _cmd_bsmp.FAC_DCDC,
        'FAC_2S_DCDC': _cmd_bsmp.FAC_2S_DCDC,
        'FAC_2S_ACDC': _cmd_bsmp.FAC_2S_ACDC,
        'FAC_2P4S_DCDC': _cmd_bsmp.FAC_2P4S_DCDC,
        'FAC_2P4S_ACDC': _cmd_bsmp.FAC_2P4S_ACDC,
        'FAP': _cmd_bsmp.FAP,
        'FAP_2P2S': _cmd_bsmp.FAP_2P2S,
        'FAP_4P': _cmd_bsmp.FAP_4P,
    }

    @staticmethod
    def create(psmodel, *args, **kwargs):
        """Return PSModel object."""
        if psmodel in PSBSMPFactory.psname_2_psbsmp:
            factory = PSBSMPFactory.psname_2_psbsmp[psmodel]
            return factory(*args, **kwargs)

        raise ValueError('PS Model "{}" not defined'.format(psmodel))



#!/usr/bin/python-sirius

import sys as _sys
import logging as _log

from ..namesys import SiriusPVName as PVName
from ..search import LLTimeSearch


_disclaimer = """
# This file was generated automatically from the data of the
# excel file Cabos_e_Fibras_Sirius.xlsx, located in:
#   https://docs.google.com/spreadsheets/d/19lNNPWxZJv5s-VTrwZRMNWLDMqdHzOQa3ZDIw5neYFI/edit?usp=sharing
#
# The function that created this file can be found in:
#   siriuspy.timesys.static_table.create_static_table.
#
# If the mentioned file change, please, run the script
# again and copy the generated file to replace this one.
"""
_NAMES2CONVERT = {
    'LA-RF:H1TRIG': 'LI-RaRF01:RF-LLRFProc',
    'LA-RF:H1SOAM-1': 'LI-RaRF02:RF-SSAmp-1',
    'LA-RF:H1SOAM-2': 'LI-RaRF02:RF-SSAmp-2',
    'LA-RF:H1SOAM-3': 'LI-RaRF02:RF-SSAmp-3',
    'LA-BI:H1FO-1': 'LI-RaDiag02:TI-TrigFout',
    'LA-MD:H1PPS-1': 'LI-RaMD01:MD-PPS',  # ?
    'LA-MD:H1PPS-2': 'LI-RaMD02:MD-PPS',  # ?
    '?': 'IA-00RaCtrl:CO-DIO',
    '"Rack" Streak Camera:TI-EVE': 'IA-00RaCtrl:TI-EVE'}


def create_static_table(fname=None, local=False, logfile=None):
    """."""
    if logfile:
        hand = _log.FileHandler(logfile, mode='w')
    else:
        hand = _log.StreamHandler(stream=_sys.stdout)
    hand.setFormatter(_log.Formatter('', datefmt=''))
    _log.getLogger().setLevel(level=_log.INFO)
    _log.getLogger().addHandler(hand)

    if local:
        data = read_data_from_local_excel_file(fname)
    else:
        data = read_data_from_google()
    _log.info(_disclaimer)
    chans = _get_channels_from_data(data)
    chans_used, chans_nused = _sort_connection_table(chans)
    _print_tables(chans_used, chans_nused)


def read_data_from_google():
    """Shows basic usage of the Sheets API.
    Prints values from a sample spreadsheet.
    """
    from googleapiclient.discovery import build
    from httplib2 import Http
    from oauth2client import file, client, tools
    _log.getLogger('googleapiclient.discovery_cache').setLevel(_log.ERROR)
    _log.getLogger('googleapiclient.discovery').setLevel(_log.ERROR)
    _log.getLogger('oauth2client.transport').setLevel(_log.ERROR)
    _log.getLogger('oauth2client.client').setLevel(_log.ERROR)

    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    store = file.Storage('/home/fernando/token.json')
    creds = store.get()
    if not creds or creds.invalid:
        flow = client.flow_from_clientsecrets(
            '/home/fernando/credentials.json',
            'https://www.googleapis.com/auth/spreadsheets.readonly')
        creds = tools.run_flow(flow, store)
    service = build('sheets', 'v4', http=creds.authorize(Http()))

    # Call the Sheets API
    sheet = service.spreadsheets()
    result = sheet.values().get(
        spreadsheetId='19lNNPWxZJv5s-VTrwZRMNWLDMqdHzOQa3ZDIw5neYFI',
        range='Cabos e Fibras').execute()
    values = result.get('values', [])

    if not values:
        raise ValueError('Error loading file from google')
    return values


def read_data_from_local_excel_file(fname=None):
    from openpyxl import load_workbook
    fname = fname or 'Cabos_e_Fibras_Sirius.xlsx'
    wb = load_workbook(fname, data_only=True)
    ws = wb['Cabos e Fibras']
    data = [[cel.value for cel in row] for row in ws.iter_rows()]
    if not data:
        raise ValueError('Error loading local file')
    return data


def _get_channels_from_data(data):
    for i, cel in enumerate(data[0]):
        if cel is None:
            continue
        val = cel.lower().replace(' ', '')
        if val.endswith('equipamento1'):
            eq1 = i
        elif val.endswith('equipamento2'):
            eq2 = i
        elif val.endswith('porta1'):
            p1 = i
        elif val.endswith('porta2'):
            p2 = i
    chans = list()
    for i, row in enumerate(data[1:]):
        if not row:
            continue
        sis = row[0]
        if sis is None or not sis.lower().startswith(('timing', 'controle')):
            continue
        name1 = _check_device_and_port(row[eq1], row[p1])
        name2 = _check_device_and_port(row[eq2], row[p2])
        try:
            name1 = PVName(name1)
            name2 = PVName(name2)
        except IndexError:
            _log.info('# {0:04d}:   {1:40s} {2:40s}'.format(i, name1, name2))
            continue
        chans.append((name1, name2))
    return chans


def _check_device_and_port(dev, por):
    por = por.upper().translate(str.maketrans('', '', ' _-'))
    return _NAMES2CONVERT.get(dev, dev) + ':' + por


def _sort_connection_table(chans):
    dev = 'EVG'
    for k1, k2 in chans:
        if k1.dev == dev:
            dev = k1
            break
        elif k2.dev == dev:
            dev = k2
            break
    else:
        raise KeyError('EVG not Found.')

    entries = LLTimeSearch.get_channel_input(
        PVName(dev.device_name+':'+'UPLINK'))
    chans_used = []
    for entry in entries:
        mark = list(range(len(chans)))
        for i, ks in enumerate(chans):
            k1, k2 = ks
            if k1 == entry:
                entries.extend(LLTimeSearch.get_channel_input(k2))
                chans_used.append((k1, k2))
                mark.remove(i)
            if k2 == entry:
                entries.extend(LLTimeSearch.get_channel_input(k1))
                chans_used.append((k2, k1))
                mark.remove(i)
        chans = [chans[i] for i in mark]
    chans_nused = sorted(chans)
    return chans_used, chans_nused


def _print_tables(chans_used, chans_nused):
    _log.info(3*'\n')
    _log.info(f'# {len(chans_used):d}')
    for k1, k2 in chans_used:
        _log.info('{0:35s} {1:35s}'.format(k1, k2))

    _log.info(5*'\n')
    _log.info('# CONNECTIONS NOT USED')
    _log.info(f'# {len(chans_nused):d}')
    for k1, k2 in chans_nused:
        _log.info('# {0:35s} {1:35s}'.format(k1, k2))



"""Siriuspy."""

import os as _os
with open(_os.path.join(__path__[0], 'VERSION'), 'r') as _f:
    __version__ = _f.read().strip()
del _os


__all__ = [
    'callbacks', 'csdev', 'envars', 'thread', 'util',
    'bsmp', 'clientarch', 'clientconfigdb', 'clientweb', 'currinfo',
    'cycle', 'devices', 'diagbeam', 'diagsys', 'epics', 'machshift',
    'magnet', 'meas', 'namesys', 'optics', 'opticscorr', 'oscilloscope',
    'posang', 'pwrsupply', 'ramp', 'search', 'simul', 'sofb', 'timesys',
    ]



"""Define Insertion Devices."""

import time as _time
import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _SiriusPVName
from ..search import IDSearch as _IDSearch
from ..magnet.idffwd import APUFFWDCalc as _APUFFWDCalc

from .device import Device as _Device
from .device import Devices as _Devices
from .device import DeviceApp as _DeviceApp
from .psconv import StrengthConv as _StrengthConv


class APU(_Device):
    """Insertion Device APU."""

    class DEVICES:
        """."""

        APU22_06SB = 'SI-06SB:ID-APU22'
        APU22_07SP = 'SI-07SP:ID-APU22'
        APU22_08SB = 'SI-08SB:ID-APU22'
        APU22_09SA = 'SI-09SA:ID-APU22'
        APU58_11SP = 'SI-11SP:ID-APU58'
        ALL = (APU22_06SB, APU22_07SP, APU22_08SB, APU22_09SA,
               APU58_11SP, )

    _properties = (
        'DevCtrl-Cmd', 'Moving-Mon',
        'PhaseSpeed-SP', 'PhaseSpeed-Mon',
        'Phase-SP', 'Phase-Mon',
        'Kx-SP', 'Kx-Mon',
    )

    _DEF_TIMEOUT = 10  # [s]
    _CMD_MOVE = 3
    _MOVECHECK_SLEEP = 0.1  # [s]

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in APU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=APU._properties, auto_mon=True)

    @property
    def phase(self):
        """Return APU phase [mm]."""
        return self['Phase-Mon']

    @phase.setter
    def phase(self, value):
        """Set APU phase [mm]."""
        self['Phase-SP'] = value

    @property
    def phase_sp(self):
        """Return APU phase SP [mm]."""
        return self['Phase-SP']

    @property
    def phase_speed(self):
        """Return APU phase speed [mm/s]."""
        return self['PhaseSpeed-Mon']

    @phase_speed.setter
    def phase_speed(self, value):
        """Set APU phase_speed [mm/s]."""
        self['PhaseSpeed-SP'] = value

    @property
    def phase_speed_sp(self):
        """Return APU phase speed SP [mm/s]."""
        return self['PhaseSpeed-SP']

    @property
    def idkx(self):
        """Return APU Kx."""
        return self['Kx-SP']

    @idkx.setter
    def idkx(self, value):
        """Set APU Kx."""
        self['Kx-SP'] = value

    @property
    def is_moving(self):
        """Return True if phase is changing."""
        return round(self['Moving-Mon']) == 1

    def cmd_move(self, timeout=_DEF_TIMEOUT):
        """."""
        self['DevCtrl-Cmd'] = APU._CMD_MOVE
        return True

    def wait_move(self):
        """Wait for phase movement to complete."""
        _time.sleep(APU._MOVECHECK_SLEEP)
        while self.is_moving:
            _time.sleep(APU._MOVECHECK_SLEEP)


class EPU(_Device):
    """."""

    class DEVICES:
        """."""

        EPU50_10SB = 'SI-10SB:ID-EPU50'
        ALL = (EPU50_10SB, )

    _idparam_fields = (
        'PERIOD',  # [mm]
        'PHASE_PARK',  # [mm]
        'GAP_PARK',  # [mm]
        )

    _dev2params = {
        DEVICES.EPU50_10SB:
            _get_namedtuple('IDParameters',
                _idparam_fields, (50.0, 0, 300.0)),
        }

    _short_shut_eye = 0.1  # [s]
    _default_timeout = 5  # [s]

    _properties = (
        'BeamLineCtrlEnbl-Sel', 'BeamLineCtrlEnbl-Sts',
        'EnblPwrAll-Cmd', 'PwrPhase-Mon', 'PwrGap-Mon',
        'EnblAndReleasePhase-Sel', 'EnblAndReleasePhase-Sts',
        'AllowedToChangePhase-Mon',
        'EnblAndReleaseGap-Sel', 'EnblAndReleaseGap-Sts',
        'AllowedToChangeGap-Mon',
        'Phase-SP', 'Phase-RB', 'Phase-Mon',
        'PhaseSpeed-SP', 'PhaseSpeed-RB', 'PhaseSpeed-Mon',
        'MaxPhaseSpeed-SP', 'MaxPhaseSpeed-RB',
        'ChangePhase-Cmd',
        'Gap-SP', 'Gap-RB', 'Gap-Mon',
        'GapSpeed-SP', 'GapSpeed-RB', 'GapSpeed-Mon',
        'MaxGapSpeed-SP', 'MaxGapSpeed-RB',
        'ChangeGap-Cmd',
        'Stop-Cmd', 'Moving-Mon', 'IsBusy-Mon',
        'Status-Mon',
        )

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in EPU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=EPU._properties, auto_mon=True)

    @property
    def parameters(self):
        """Return EPU parameters."""
        return EPU._dev2params[self.devname]

    @property
    def status(self):
        """EPU status."""
        return self['Status-Mon']

    # --- speeds ----

    @property
    def phase_speed(self):
        """Return phase speed readback [mm/s]."""
        return self['PhaseSpeed-RB']

    @property
    def phase_speed_mon(self):
        """Return phase speed monitor [mm/s]."""
        return self['PhaseSpeed-Mon']

    @property
    def phase_speed_max(self):
        """Return max phase speed readback [mm/s]."""
        return self['MaxPhaseSpeed-RB']

    @property
    def gap_speed(self):
        """Return gap speed readback [mm/s]."""
        return self['GapSpeed-RB']

    @property
    def gap_speed_mon(self):
        """Return gap speed monitor [mm/s]."""
        return self['GapSpeed-Mon']

    @property
    def gap_speed_max(self):
        """Return max gap speed readback [mm/s]."""
        return self['MaxGapSpeed-RB']

    # --- phase ---

    @property
    def phase(self):
        """Return EPU phase readback [mm]."""
        return self['Phase-RB']

    @property
    def phase_min(self):
        """Return EPU phase lower control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Phase-SP')
        return ctrlvars['lower_ctrl_limit']

    @property
    def phase_max(self):
        """Return EPU phase upper control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Phase-SP')
        return ctrlvars['upper_ctrl_limit']

    @property
    def phase_mon(self):
        """Return EPU phase monitor [mm]."""
        return self['Phase-Mon']

    # --- gap ---

    @property
    def gap(self):
        """Return EPU gap readback [mm]."""
        return self['Gap-RB']

    @property
    def gap_min(self):
        """Return EPU gap lower control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Gap-SP')
        return ctrlvars['lower_ctrl_limit']

    @property
    def gap_max(self):
        """Return EPU gap upper control limit [mm]."""
        ctrlvars = self.pv_ctrlvars('Gap-SP')
        return ctrlvars['upper_ctrl_limit']

    @property
    def gap_mon(self):
        """Return EPU gap monitor [mm]."""
        return self['Gap-Mon']

    # --- drive checks ---

    @property
    def is_phase_drive_powered(self):
        """Return phase driver power state on (True|False)."""
        return self['PwrPhase-Mon'] != 0

    @property
    def is_gap_drive_powered(self):
        """Return gap driver power state on (True|False)."""
        return self['PwrGap-Mon'] != 0

    @property
    def is_drives_powered(self):
        """Return phase & gap drives powered state on (True|False)."""
        return self.is_phase_drive_powered and self.is_gap_drive_powered

    # --- movement checks ---

    @property
    def is_move_phase_enabled(self):
        """Return phase movement enabled state (True|False)."""
        return self['AllowedToChangePhase-Mon'] != 0

    @property
    def is_move_gap_enabled(self):
        """Return gap movement enabled state (True|False)."""
        return self['AllowedToChangeGap-Mon'] != 0

    @property
    def is_move_enabled(self):
        """Return phase and gap movements enabled state (True|False)."""
        return self.is_move_phase_enabled and self.is_move_gap_enabled

    @property
    def is_moving(self):
        """Return is moving state (True|False)."""
        return self['Moving-Mon'] != 0

    # --- other checks ---

    @property
    def is_busy(self):
        """Return is busy state (True|False)."""
        return self['IsBusy-Mon'] != 0

    @property
    def is_beamline_ctrl_enabled(self):
        """Return beamline control enabled state (True|False)."""
        return self['BeamLineCtrlEnbl-Sts'] != 0

    # --- cmd_wait

    def cmd_wait_while_busy(self, timeout=None):
        """Command wait within timeout while EPU control is busy."""
        timeout = timeout or self._default_timeout
        time_init = _time.time()
        while self.is_busy:
            _time.sleep(min(EPU._short_shut_eye, timeout))
            if _time.time() - time_init > timeout:
                return False
        return True

    # --- cmd_beamline and cmd_drive

    def cmd_drive_turn_on(self, timeout=None):
        """Command turn phase and gap drives on."""
        if self.is_phase_drive_powered and self.is_gap_drive_powered:
            return True
        self['EnblPwrAll-Cmd'] = 1
        props_values = {
            'PwrPhase-Mon': 0,  # value compared to in 'not equal' operation
            'PwrGap-Mon': 0,  # value compared to in 'not equal' operation
            }
        return self._wait(props_values, timeout=timeout, comp='ne')

    def cmd_beamline_ctrl_enable(self, timeout=None):
        """Command enable bealine EPU control."""
        return self._set_sp('BeamLineCtrlEnbl-Sel', 1, timeout)

    def cmd_beamline_ctrl_disable(self, timeout=None):
        """Command disable bealine EPU control."""
        return self._set_sp('BeamLineCtrlEnbl-Sel', 0, timeout)

    # --- cmd_set ---

    def cmd_set_phase(self, phase, timeout=None):
        """Command to set EPU target phase for movement [mm]."""
        return self._set_sp('Phase-SP', phase, timeout)

    def cmd_set_gap(self, gap, timeout=None):
        """Command to set EPU target gap for movement [mm]."""
        return self._set_sp('Gap-SP', gap, timeout)

    def cmd_set_phase_speed(self, phase_speed, timeout=None):
        """Command to set EPU cruise phase speed for movement [mm/s]."""
        return self._set_sp('PhaseSpeed-SP', phase_speed, timeout)

    def cmd_set_gap_speed(self, gap_speed, timeout=None):
        """Command to set EPU cruise gap speed for movement [mm/s]."""
        return self._set_sp('GapSpeed-SP', gap_speed, timeout)

    def cmd_set_phase_speed_max(self, phase_speed_max, timeout=None):
        """Command to set EPU max cruise phase speed for movement [mm/s]."""
        return self._set_sp('MaxPhaseSpeed-SP', phase_speed_max, timeout)

    def cmd_set_gap_speed_max(self, gap_speed_max, timeout=None):
        """Command to set EPU max cruise gap speed for movement [mm/s]."""
        return self._set_sp('MaxGapSpeed-SP', gap_speed_max, timeout)

    # --- cmd_move disable/enable ---

    def cmd_move_phase_enable(self, timeout=None):
        """Command to release and enable EPU phase movement."""
        # return self._set_sp('EnblAndReleasePhase-Sel', 1, timeout)
        self['EnblAndReleasePhase-Sel'] = 1
        return True

    def cmd_move_phase_disable(self, timeout=None):
        """Command to disable and break EPU phase movement."""
        # return self._set_sp('EnblAndReleasePhase-Sel', 0, timeout)
        self['EnblAndReleasePhase-Sel'] = 0
        return True

    def cmd_move_gap_enable(self, timeout=None):
        """Command to release and enable EPU gap movement."""
        # return self._set_sp('EnblAndReleaseGap-Sel', 1, timeout)
        self['EnblAndReleaseGap-Sel'] = 1
        return True

    def cmd_move_gap_disable(self, timeout=None):
        """Command to disable and break EPU gap movement."""
        # return self._set_sp('EnblAndReleaseGap-Sel', 0, timeout)
        self['EnblAndReleaseGap-Sel'] = 0
        return True

    def cmd_move_enable(self, timeout=None):
        """Command to release and enable EPU phase and gap movements."""
        success = True
        success &= self.cmd_move_phase_enable(timeout=timeout)
        success &= self.cmd_move_gap_enable(timeout=timeout)
        return success

    def cmd_move_disable(self, timeout=None):
        """Command to disable and break EPU phase and gap movements."""
        success = True
        success &= self.cmd_move_phase_disable(timeout=timeout)
        success &= self.cmd_move_gap_disable(timeout=timeout)
        return success

    # -- cmd_move

    def cmd_move_stop(self, timeout=None):
        """Command to interrupt and then enable phase and gap movements."""
        timeout = timeout or self._default_timeout

        # wait for not busy state
        if self.cmd_wait_while_busy(timeout=timeout):
            return False

        # send stop command
        self['Stop-Cmd'] = 1

        # check for successful stop
        if self.cmd_wait_while_busy(timeout=timeout):
            return False
        success = True
        success &= super()._wait('Moving-Mon', 0, tiemout=timeout)
        success &= super()._wait('IsBusy-Mon', 0, tiemout=timeout)
        if not success:
            return False

        # enable movement again
        return self.cmd_move_enable(timeout=timeout)

    def cmd_move_phase_start(self, timeout=None):
        """Command to start phase movement."""
        return self._move_start('ChangePhase-Cmd', timeout=timeout)

    def cmd_move_gap_start(self, timeout=None):
        """Command to start gap movement."""
        return self._move_start('ChangeGap-Cmd', timeout=timeout)

    def cmd_move(self, phase, gap, timeout=None):
        """Command to set and start phase and gap movements."""
        # calc ETA
        dtime_phase = abs(phase - self.phase_mon) / self.phase_speed
        dtime_gap = abs(gap - self.gap_mon) / self.gap_speed
        dtime_max = max(dtime_phase, dtime_gap)

        # additional percentual in ETA
        tol_gap = 0.002  # [mm]
        tol_phase = 0.002  # [mm]
        tol_dtime = 100  # [%]
        tol_factor = (1 + tol_dtime/100)
        tol_total = tol_factor * dtime_max + 5

        # set target phase and gap
        if not self.cmd_set_phase(phase=phase, timeout=timeout):
            return False
        if not self.cmd_set_gap(gap=gap, timeout=timeout):
            return False

        # command move start
        if not self.cmd_move_phase_start(timeout=timeout):
            return False
        if not self.cmd_move_gap_start(timeout=timeout):
            return False

        # wait for movement within reasonable time
        time_init = _time.time()
        while \
                abs(self.gap_mon - gap) > tol_gap or \
                abs(self.phase_mon - phase) > tol_phase or \
                self.is_moving:
            if _time.time() - time_init > tol_total:
                print(f'tol_total: {tol_total:.3f} s')
                print(f'wait_time: {_time.time() - time_init:.3f} s')
                print()
                return False
            _time.sleep(EPU._short_shut_eye)

        # successfull movement at this point
        return True

    def cmd_move_park(self, timeout=None):
        """Command to set and start EPU movement to parked config."""
        params = self.parameters
        return self.cmd_move(
            params.PHASE_PARK, params.GAP_PARK, timeout=timeout)

    # --- cmd_reset

    def cmd_device_reset(self, timeout=None):
        """Command to reset EPU to a standard movement state."""
        success = True
        success &= self.cmd_beamline_ctrl_disable(timeout=timeout)
        success &= self.cmd_drive_turn_on(timeout=timeout)
        success &= self.cmd_move_enable(timeout=timeout)
        return success

    # --- private methods ---

    def _move_start(self, cmd_propty, timeout=None):
        """."""
        timeout = timeout or self._default_timeout

        # wait for not busy state
        if not self.cmd_wait_while_busy(timeout=timeout):
            return False

        # send move command
        self[cmd_propty] = 1

        return True

    def _set_sp(self, propties_sp, values, timeout=None):
        timeout = timeout or self._default_timeout
        if isinstance(propties_sp, str):
            propties_sp = (propties_sp, )
            values = (values, )
        success = True
        for propty_sp, value in zip(propties_sp, values):
            if value == self[propty_sp]:
                continue
            if not self.cmd_wait_while_busy(timeout=timeout):
                return False
            self[propty_sp] = value
            propty_rb = propty_sp.replace('-SP', '-RB').replace('-Sel', '-Sts')
            success &= super()._wait(
                propty_rb, value, timeout=timeout, comp='eq')
        return success

    def _wait(self, props_values, timeout=None, comp='eq'):
        timeout = timeout or self._default_timeout
        success = True
        for propty, value in props_values.items():
            success &= super()._wait(
                propty, value, timeout=timeout, comp=comp)
        return success


class IDCorrectors(_DeviceApp):
    """."""

    DEVICES = APU.DEVICES

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in IDCorrectors.DEVICES.ALL:
            raise NotImplementedError(devname)

        # get correctors names
        self._psnames_orb = _IDSearch.conv_idname_2_orbitcorr(devname)

        # get deviceapp properties
        properties, \
            self._orb_sp, self._orb_rb, self._orb_refmon, self._orb_mon = \
            self._get_properties()

        # call base class constructor
        super().__init__(properties=properties, devname=devname)

    @property
    def orbitcorr_psnames(self):
        """Return orbit corrector names."""
        return self._psnames_orb

    @property
    def orbitcorr_current(self):
        """Return orbit SOFBCurrent Mon."""
        return self[self._orb_mon]

    @orbitcorr_current.setter
    def orbitcorr_current(self, value):
        """Set orbit SOFBCurrent SP."""
        self[self._orb_sp] = value

    @property
    def orbitcorr_current_sp(self):
        """Return orbit SOFBCurrent setpoint."""
        return self[self._orb_sp]

    @property
    def orbitcorr_current_rb(self):
        """Return orbit SOFBCurrent readback."""
        return self[self._orb_rb]

    @property
    def orbitcorr_current_mon(self):
        """Return orbit SOFBCurrent monitor."""
        return self[self._orb_mon]

    def _get_properties(self):
        corrname = self._psnames_orb[0]
        orb_sp = corrname + ':SOFBCurrent-SP'
        orb_rb = corrname + ':SOFBCurrent-RB'
        orb_refmon = corrname + ':SOFBCurrentRef-Mon'
        orb_mon = corrname + ':SOFBCurrent-Mon'
        properties = (
            orb_sp, orb_rb, orb_refmon, orb_mon
        )
        return properties, orb_sp, orb_rb, orb_refmon, orb_mon


class APUFeedForward(_Devices):
    """Insertion Device APU FeedForward."""

    DEVICES = APU.DEVICES

    def __init__(self, devname):
        """."""
        # create APU device
        self._apu = APU(devname)

        # create IDCorrectors
        self._idcorrs = IDCorrectors(devname)

        # create FFWDCalc
        self._ffwdcalc = _APUFFWDCalc(devname)

        # create normalizers
        self._strenconv_chs, self._strenconv_cvs = self._create_strenconv()

        # call base class constructor
        devices = (
            self._apu, self._idcorrs,
            self._strenconv_chs, self._strenconv_cvs)
        super().__init__(devname, devices)

        # bumps
        self._posx, self._angx, self._posy, self._angy = \
            self._init_posang()

    @property
    def apu(self):
        """Return APU device."""
        return self._apu

    @property
    def correctors(self):
        """Return IDCorrectors device."""
        return self._idcorrs

    @property
    def ffwdcalc(self):
        """."""
        return self._ffwdcalc

    @property
    def posx(self):
        """Return posx bump value."""
        return self._posx

    @posx.setter
    def posx(self, value):
        """Return posx bump value."""
        self._posx = value

    @property
    def angx(self):
        """Return angx bump value."""
        return self._angx

    @angx.setter
    def angx(self, value):
        """Return angx bump value."""
        self._angx = value

    @property
    def posy(self):
        """Return posy bump value."""
        return self._posy

    @posy.setter
    def posy(self, value):
        """Return posy bump value."""
        self._posy = value

    @property
    def angy(self):
        """Return angy bump value."""
        return self._angy

    @angy.setter
    def angy(self, value):
        """Return angy bump value."""
        self._angy = value

    def conv_orbitcorr_kick2curr(self, kicks):
        """."""
        # 127 µs ± 1.01 µs per loop
        kickx = kicks[:self.ffwdcalc.nr_chs]
        kicky = kicks[self.ffwdcalc.nr_chs:]
        curr_chs = self._strenconv_chs.conv_strength_2_current(kickx)
        curr_cvs = self._strenconv_cvs.conv_strength_2_current(kicky)
        currs = _np.hstack((curr_chs, curr_cvs))
        return currs

    def bump_get_orbitcorr_current(self):
        """Return bump orbit correctors currents."""
        kicks = self.ffwdcalc.conv_posang2kick(
            self.posx, self.angx, self.posy, self.angy)
        currents = self.conv_orbitcorr_kick2curr(kicks)
        return currents

    def ffwd_get_orbitcorr_current(self, phase=None):
        """Return feedforward orbitcorr currents for a given ID phase."""
        # 157 µs ± 3.93 µs per loop [if phase is passed]
        if phase is None:
            phase = self.apu.phase
        currents = self.ffwdcalc.conv_phase_2_orbcorr_currents(phase)
        return currents

    def ffwd_update_orbitcorr(self, phase=None):
        """Update orbit feedforward."""
        currents_ffwd = self.ffwd_get_orbitcorr_current(phase)
        currents_bump = self.bump_get_orbitcorr_current()
        self.correctors.orbitcorr_current = currents_ffwd + currents_bump

    def ffwd_update(self, phase=None):
        """Update feedforward with bump."""
        # 305 µs ± 45.5 µs per loop
        self.ffwd_update_orbitcorr(phase)

    # --- private methods ---

    def _init_posang(self):
        _ = self  # throaway arguments (temporary)
        posx, angx = 0.0, 0.0
        posy, angy = 0.0, 0.0
        # NOTE: we could initialize posang with corrector values.
        return posx, angx, posy, angy

    def _create_strenconv(self):
        """."""
        psnames = self.correctors.orbitcorr_psnames

        maname = psnames[0].replace(':PS-', ':MA-')
        strenconv_chs = _StrengthConv(
            maname, proptype='Ref-Mon', auto_mon=True)
        maname = psnames[self.ffwdcalc.nr_chs].replace(':PS-', ':MA-')
        strenconv_cvs = _StrengthConv(
            maname, proptype='Ref-Mon', auto_mon=True)

        return strenconv_chs, strenconv_cvs



"""PV Arch Module."""

from copy import deepcopy as _dcopy

import numpy as _np

from mathphys.functions import save_pickle as _save_pickle, \
    load_pickle as _load_pickle

from . import exceptions as _exceptions
from .client import ClientArchiver as _ClientArchiver
from .time import Time as _Time, get_time_intervals as _get_time_intervals


class _Base:

    def __init__(self, connector=None):
        self._connector = None
        self.connector = connector
        self.connect()

    @property
    def is_archived(self):
        """Is archived."""
        self.connect()
        return self.connector.getPVDetails(self.pvname) is not None

    def connect(self):
        """Connect."""
        if self.connector is None:
            self._connector = _ClientArchiver()

    @property
    def connector(self):
        """Connector."""
        return self._connector

    @connector.setter
    def connector(self, conn):
        if conn is None:
            return
        elif isinstance(conn, _ClientArchiver):
            self._connector = conn
        elif isinstance(conn, str):
            self._connector = _ClientArchiver(server_url=conn)
        else:
            raise TypeError(
                'Variable conn must be a str or ClientArchiver object.')

    @property
    def timeout(self):
        """Connection timeout."""
        return self.connector.timeout

    @timeout.setter
    def timeout(self, value):
        """Set connection timeout."""
        self.connector.timeout = float(value)

    @property
    def connected(self):
        """."""
        if not self.connector:
            return False
        return self.connector.connected


class PVDetails(_Base):
    """Archive PV Details."""

    _field2type = {
        'Number of elements': ('nelms', int),
        'Units:': ('units', str),
        'Host name': ('host_name', str),
        'Average bytes per event': ('avg_bytes_per_event', float),
        'Estimated storage rate (KB/hour)':
            ('estimated_storage_rate_kb_hour', float),
        'Estimated storage rate (MB/day)':
            ('estimated_storage_rate_mb_day', float),
        'Estimated storage rate (GB/year)':
            ('estimated_storage_rate_gb_year', float),
        }

    def __init__(self, pvname, connector=None):
        """."""
        super().__init__(connector)
        self.pvname = pvname
        self.is_scalar = None
        self.is_paused = None
        self.is_connected = None
        self.nelms = None
        self.units = None
        self.host_name = None
        self.avg_bytes_per_event = None
        self.estimated_storage_rate_kb_hour = None
        self.estimated_storage_rate_mb_day = None
        self.estimated_storage_rate_gb_year = None

    @property
    def request_url(self):
        """."""
        self.connect()
        url = self.connector.getPVDetails(self.pvname, get_request_url=True)
        return url

    @property
    def is_archived(self):
        """."""
        self.connect()
        data = self.connector.getPVDetails(self.pvname)
        if not data:
            return False
        return True

    def update(self):
        """."""
        self.connect()
        data = self.connector.getPVDetails(self.pvname)
        if not data:
            return False
        for datum in data:
            # print(datum)
            field, value = datum['name'], datum['value']
            if value is None:
                continue
            # value = value.replace(',', '.')
            value = value.replace(',', '')
            if field in PVDetails._field2type:
                fattr, ftype = PVDetails._field2type[field]
                if not value == 'Not enough info':
                    value = ftype(value)
                setattr(self, fattr, value)
            elif field == 'Is this a scalar:':
                self.is_scalar = (value.lower() == 'yes')
            elif field == 'Is this PV paused:':
                self.is_paused = (value.lower() == 'yes')
            elif field == 'Is this PV currently connected?':
                self.is_connected = (value.lower() == 'yes')
        return True

    def __str__(self):
        """."""
        rst = ''
        rst += '{:<30s}: {:}\n'.format('pvname', self.pvname)
        rst += '{:<30s}: {:}\n'.format('is_scalar', self.is_scalar)
        rst += '{:<30s}: {:}\n'.format('is_paused', self.is_paused)
        rst += '{:<30s}: {:}\n'.format('is_connected', self.is_connected)
        rst += '{:<30s}: {:}\n'.format('nelms', self.nelms)
        rst += '{:<30s}: {:}\n'.format('units', self.units)
        rst += '{:<30s}: {:}\n'.format('host_name', self.host_name)
        rst += '{:<30s}: {:}\n'.format(
            'avg_bytes_per_event', self.avg_bytes_per_event)
        rst += '{:<30s}: {:}\n'.format(
            'estimated_storage_rate_kb_hour',
            self.estimated_storage_rate_kb_hour)
        rst += '{:<30s}: {:}\n'.format(
            'estimated_storage_rate_mb_day',
            self.estimated_storage_rate_mb_day)
        rst += '{:<30s}: {:}\n'.format(
            'estimated_storage_rate_gb_year',
            self.estimated_storage_rate_gb_year)
        return rst


class PVData(_Base):
    """Archive PV Data."""

    def __init__(self, pvname, connector=None):
        """Initialize."""
        super().__init__(connector)
        self._pvname = pvname
        self._time_start = None
        self._time_stop = None
        self._timestamp = None
        self._value = None
        self._status = None
        self._severity = None
        self._parallel_query_bin_interval = 12*60*60  # 12h

    @property
    def pvname(self):
        """PVName."""
        return self._pvname

    @property
    def request_url(self):
        """Request url."""
        self.connect()
        url = self.connector.getData(
            self.pvname,
            self._time_start.get_iso8601(),
            self._time_stop.get_iso8601(),
            get_request_url=True)
        return url

    @property
    def timestamp_start(self):
        """Timestamp start."""
        if not self._time_start:
            return None
        return self._time_start.timestamp()

    @timestamp_start.setter
    def timestamp_start(self, new_timestamp):
        if not isinstance(new_timestamp, (float, int)):
            raise _exceptions.TypeError(
                'expected argument of type float or int, got ' +
                str(type(new_timestamp)))
        self._time_start = _Time(timestamp=new_timestamp)

    @property
    def time_start(self):
        """Time start."""
        return self._time_start

    @time_start.setter
    def time_start(self, new_time):
        if not isinstance(new_time, _Time):
            raise _exceptions.TypeError(
                'expected argument of type Time, got '+str(type(new_time)))
        self._time_start = new_time

    @property
    def timestamp_stop(self):
        """Timestamp stop."""
        if not self._time_stop:
            return None
        return self._time_stop.timestamp()

    @timestamp_stop.setter
    def timestamp_stop(self, new_timestamp):
        if not isinstance(new_timestamp, (float, int)):
            raise _exceptions.TypeError(
                'expected argument of type float or int, got ' +
                str(type(new_timestamp)))
        self._time_stop = _Time(timestamp=new_timestamp)

    @property
    def time_stop(self):
        """Time stop."""
        return self._time_stop

    @time_stop.setter
    def time_stop(self, new_time):
        if not isinstance(new_time, _Time):
            raise _exceptions.TypeError(
                'expected argument of type Time, got ' + str(type(new_time)))
        self._time_stop = new_time

    @property
    def parallel_query_bin_interval(self):
        """Parallel query bin interval."""
        return self._parallel_query_bin_interval

    @parallel_query_bin_interval.setter
    def parallel_query_bin_interval(self, new_intvl):
        if not isinstance(new_intvl, (float, int)):
            raise _exceptions.TypeError(
                'expected argument of type float or int, got ' +
                str(type(new_intvl)))
        self._parallel_query_bin_interval = new_intvl

    @property
    def timestamp(self):
        """Timestamp data."""
        return self._timestamp

    @property
    def value(self):
        """Value data."""
        return self._value

    @property
    def status(self):
        """Status data."""
        return self._status

    @property
    def severity(self):
        """Severity data."""
        return self._severity

    def update(self, mean_sec=None, parallel=True):
        """Update."""
        self.connect()
        if None in (self.timestamp_start, self.timestamp_stop):
            print('Start and stop timestamps not defined! Aborting.')
            return
        process_type = 'mean' if mean_sec is not None else ''

        interval = self.parallel_query_bin_interval
        if parallel:
            timestamp_start, timestamp_stop = _get_time_intervals(
                self._time_start, self._time_stop, interval,
                return_isoformat=True)
        else:
            timestamp_start = self._time_start.get_iso8601()
            timestamp_stop = self._time_stop.get_iso8601()

        data = self.connector.getData(
            self._pvname, timestamp_start, timestamp_stop,
            process_type=process_type, interval=mean_sec)
        if not data:
            return
        self.set_data(**data)

    def set_data(self, timestamp, value, status, severity):
        """Auxiliary method to set data. Used by PVDataSet."""
        self._timestamp = self._value = self._status = self._severity = None
        if timestamp is not None:
            self._timestamp = _np.asarray(timestamp)
            self._value = _np.asarray(value)
            self._status = _np.asarray(status)
            self._severity = _np.asarray(severity)

    def to_dict(self):
        """Return dictionary with PV properties.

        Returns:
            dict: dictionary with fields:
                server_url (str): url of the Archiver server.
                pvname (str): the name of the PV.
                timestamp_start (time.time): start of acquisition time.
                timestamp_stop (time.time): end of acquisition time.
                data (dict): dictionary with archiver data  with fields:
                    value (numpy.ndarray): values of the PV.
                    timestamp (numpy.ndarray): timestamps of the PV.
                    status (numpy.ndarray): status of the PV.
                    severity (numpy.ndarray): severity of the PV.

        """
        return dict(
            server_url=self.connector.server_url,
            pvname=self.pvname,
            timestamp_start=self.timestamp_start,
            timestamp_stop=self.timestamp_stop,
            data=dict(
                timestamp=self.timestamp, value=self.value,
                status=self.status, severity=self.severity)
            )

    @staticmethod
    def from_dict(infos):
        """Return PVData object with information in data.

        Args:
            infos (dict): must have the same fields as the dictionary returned
                by PVData.to_dict method.

        Returns:
            PVData: with values compatible with data.

        """
        pvdata = PVData(infos['pvname'], connector=infos['server_url'])
        pvdata.timestamp_start = infos['timestamp_start']
        pvdata.timestamp_stop = infos['timestamp_stop']
        pvdata.set_data(**infos['data'])
        return pvdata

    def to_pickle(self, fname, overwrite=False):
        """Create pickle file with complete PVData.

        The data saved is the dictionary returned by PVData.to_dict method.

        Args:
            fname (str): name of the file to save.
            overwrite (bool, optional): Whether to overwrite existing file
                with same name or not. Defaults to False.

        """
        _save_pickle(self.to_dict(), fname, overwrite=overwrite)

    @staticmethod
    def from_pickle(fname):
        """Load data from pickle file and return PVData object.

        Args:
            fname (str): name of the file to load.

        Returns:
            PVData: with values from the pickle file.

        """
        return PVData.from_dict(_load_pickle(fname))


class PVDataSet(_Base):
    """A set of PVData objects."""

    def __init__(self, pvnames, connector=None):
        """Initialize."""
        super().__init__(connector)
        self._pvnames = pvnames
        self._time_start = None
        self._time_stop = None
        self._parallel_query_bin_interval = 12*60*60  # 12h
        self._pvdata = self._init_pvdatas(pvnames, self.connector)

    @property
    def pvnames(self):
        """PV names."""
        return _dcopy(self._pvnames)

    @property
    def is_archived(self):
        """Is archived."""
        self.connect()
        for pvn in self._pvnames:
            if self.connector.getPVDetails(pvn) is None:
                return False
        return True

    @property
    def timestamp_start(self):
        """Timestamp start."""
        if not self._time_start:
            return None
        return self._time_start.timestamp()

    @timestamp_start.setter
    def timestamp_start(self, new_timestamp):
        if not isinstance(new_timestamp, (float, int)):
            raise _exceptions.TypeError(
                'expected argument of type float or int, got ' +
                str(type(new_timestamp)))
        self._time_start = _Time(timestamp=new_timestamp)
        for pvname in self._pvnames:
            self._pvdata[pvname].time_start = self._time_start

    @property
    def time_start(self):
        """Time start."""
        return self._time_start

    @time_start.setter
    def time_start(self, new_time):
        if not isinstance(new_time, _Time):
            raise _exceptions.TypeError(
                'expected argument of type Time, got '+str(type(new_time)))
        self._time_start = new_time
        for pvname in self._pvnames:
            self._pvdata[pvname].time_start = self._time_start

    @property
    def timestamp_stop(self):
        """Timestamp stop."""
        if not self._time_stop:
            return None
        return self._time_stop.timestamp()

    @timestamp_stop.setter
    def timestamp_stop(self, new_timestamp):
        if not isinstance(new_timestamp, (float, int)):
            raise _exceptions.TypeError(
                'expected argument of type float or int, got ' +
                str(type(new_timestamp)))
        self._time_stop = _Time(timestamp=new_timestamp)
        for pvname in self._pvnames:
            self._pvdata[pvname].time_stop = self._time_stop

    @property
    def time_stop(self):
        """Time stop."""
        return self._time_stop

    @time_stop.setter
    def time_stop(self, new_time):
        if not isinstance(new_time, _Time):
            raise _exceptions.TypeError(
                'expected argument of type Time, got '+str(type(new_time)))
        self._time_stop = new_time
        for pvname in self._pvnames:
            self._pvdata[pvname].time_stop = self._time_stop

    @property
    def parallel_query_bin_interval(self):
        """Parallel query bin interval."""
        return self._parallel_query_bin_interval

    @parallel_query_bin_interval.setter
    def parallel_query_bin_interval(self, new_intvl):
        if not isinstance(new_intvl, (float, int)):
            raise _exceptions.TypeError(
                'expected argument of type float or int, got ' +
                str(type(new_intvl)))
        self._parallel_query_bin_interval = new_intvl
        for pvname in self._pvnames:
            self._pvdata[pvname].parallel_query_bin_interval = \
                self._parallel_query_bin_interval

    def update(self, mean_sec=None, parallel=True):
        """Update."""
        self.connect()
        if None in (self.timestamp_start, self.timestamp_stop):
            print('Start and stop timestamps not defined! Aborting.')
            return
        process_type = 'mean' if mean_sec is not None else ''

        interval = self.parallel_query_bin_interval
        if parallel:
            timestamp_start, timestamp_stop = _get_time_intervals(
                self._time_start, self._time_stop, interval,
                return_isoformat=True)
        else:
            timestamp_start = self._time_start.get_iso8601()
            timestamp_stop = self._time_stop.get_iso8601()

        data = self.connector.getData(
            self._pvnames, timestamp_start, timestamp_stop,
            process_type=process_type, interval=mean_sec)
        if not data:
            return
        for pvname in self._pvnames:
            self._pvdata[pvname].set_data(**data[pvname])

    @staticmethod
    def _init_pvdatas(pvnames, connector):
        pvdata = dict()
        for pvname in pvnames:
            pvdata[pvname] = PVData(pvname, connector)
        return pvdata

    def __getitem__(self, val):
        """Get item."""
        if isinstance(val, str):
            return self._pvdata[val]
        elif isinstance(val, int):
            return self._pvdata[self._pvnames[val]]
        else:
            raise TypeError('Item must be int or str.')

    def __iter__(self):
        """Iterate over object."""
        for pvn in self.pvnames:
            yield self[pvn]

    def __len__(self):
        """Size of the object."""
        return len(self.pvnames)

    def to_dict(self):
        """Return dictionary with PV properties.

        Returns:
            dict: dictionary with fields:
                server_url (str): url of the Archiver server.
                pvnames (list): the names of the PVs.
                timestamp_start (time.time): start of acquisition time.
                timestamp_stop (time.time): end of acquisition time.
                pvdata_info (list): with dictionaries containing the data for
                    all PVs. Compatible input for PVData.from_dict.

        """
        data = dict(
            server_url=self.connector.server_url,
            pvnames=self.pvnames,
            timestamp_start=self.timestamp_start,
            timestamp_stop=self.timestamp_stop)
        data['pvdata_info'] = [self[pvn].to_dict() for pvn in self._pvnames]
        return data

    @staticmethod
    def from_dict(info):
        """Return PVDataSet object with information in data.

        Args:
            info (dict): must have the same fields as the dictionary returned
                by PVDataSet.to_dict method.

        Returns:
            PVDataSet: with values compatible with data.

        """
        pvdataset = PVDataSet(info['pvnames'], info['server_url'])
        pvdataset.timestamp_start = info['timestamp_start']
        pvdataset.timestamp_stop = info['timestamp_stop']
        for i, pvdata in enumerate(pvdataset):
            pvdata.set_data(**info['pvdata_info'][i]['data'])
        return pvdataset

    def to_pickle(self, fname, overwrite=False):
        """Create pickle file with complete PVDataSet.

        The data saved is the dictionary returned by PVDataSet.to_dict method.

        Args:
            fname (str): name of the file to save.
            overwrite (bool, optional): Whether to overwrite existing file
                with same name or not. Defaults to False.

        """
        _save_pickle(self.to_dict(), fname, overwrite=overwrite)

    @staticmethod
    def from_pickle(fname):
        """Load data from pickle file and return PVDataSet object.

        Args:
            fname (str): name of the file to load.

        Returns:
            PVDataSet: with values from the pickle file.

        """
        return PVDataSet.from_dict(_load_pickle(fname))



"""Reader Classes.

Define classes whose objects are mapped from epics PVs to
power supply controllers properties or functions. These classes
implement a common interface that exposes the `read` method.
"""
import re as _re

from .pscstatus import PSCStatus as _PSCStatus


class Variable:
    """Readable BSMP variable."""

    def __init__(self, pru_controller, device_id, bsmp_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id
        self.bsmp_id = bsmp_id

    def read(self):
        """Read variable from pru controller."""
        value = self.pru_controller.read_variables(
            self.device_id, self.bsmp_id)
        return value


class WfmRBCurve:
    """BSMP PS WfmRB Curve read."""

    def __init__(self, pru_controller, device_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id

    def read(self):
        """Read curve."""
        data = self.pru_controller.wfmref_rb_read(self.device_id)
        return data


class WfmRefMonCurve:
    """BSMP PS WfmRef-Mon Curve read."""

    def __init__(self, pru_controller, device_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id

    def read(self):
        """Read curve."""
        data = self.pru_controller.wfmref_read(self.device_id)
        return data


class WfmMonCurve:
    """BSMP PS Wfm-Mon Curve read."""

    def __init__(self, pru_controller, device_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id

    def read(self):
        """Read curve."""
        data = self.pru_controller.scope_read(self.device_id)
        # print(self.device_id, data)
        return data


class WfmIndexCurve:
    """BSMP PS WfmIndex Curve read."""

    def __init__(self, pru_controller, device_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id

    def read(self):
        """Read curve."""
        data = self.pru_controller.wfmref_index(self.device_id)
        return data


class WfmUpdateAutoSts:
    """Wfm Update Auto Status."""

    def __init__(self, pru_controller, device_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id

    def read(self):
        """Read status."""
        data = self.pru_controller.scope_update_auto
        return data


class TimestampUpdate:
    """Timestamp update read."""

    def __init__(self, pru_controller, device_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id

    def read(self):
        """Read curve."""
        timestamp = self.pru_controller.timestamp_update()
        return timestamp


class PRUCProperty:
    """Read a PRU property."""

    def __init__(self, pru_controller, pru_property):
        """Get pru controller and property name."""
        self.pru_controller = pru_controller
        self.property = pru_property

    def read(self):
        """Read pru property."""
        return getattr(self.pru_controller, self.property)


class PwrState:
    """Variable decorator."""

    def __init__(self, variable):
        """Set variable."""
        self.variable = variable
        self.psc_status = _PSCStatus()

    def read(self):
        """Decorate read."""
        value = self.variable.read()
        if value is None:
            return value
        self.psc_status.ps_status = value
        return self.psc_status.ioc_pwrstate


class OpMode:
    """Variable decorator."""

    def __init__(self, variable):
        """Set variable."""
        self.variable = variable
        self.psc_status = _PSCStatus()

    def read(self):
        """Decorate read."""
        value = self.variable.read()
        if value is None:
            return None
        self.psc_status.ps_status = value
        return self.psc_status.state


class CtrlMode:
    """Variable decorator."""

    def __init__(self, variable):
        """Set variable."""
        self.variable = variable
        self.psc_status = _PSCStatus()

    def read(self):
        """Decorate read."""
        value = self.variable.read()
        if value is None:
            return value
        self.psc_status.ps_status = value
        return self.psc_status.interface


class CtrlLoop:
    """Variable decorator."""

    def __init__(self, variable):
        """Set variable."""
        self.variable = variable
        self.psc_status = _PSCStatus()

    def read(self):
        """Decorate read."""
        value = self.variable.read()
        if value is None:
            return value
        self.psc_status.ps_status = value
        return self.psc_status.open_loop


class Version:
    """Version variable."""

    def __init__(self, variable):
        """Set variable."""
        self.variable = variable

    def read(self):
        """Decorate read."""
        value = self.variable.read()
        if value is None:
            return None
        version = ''.join([c.decode() for c in value])
        try:
            value, _ = version.split('\x00', 0)
        except ValueError:
            value = version
        return value


class Constant:
    """Constant."""

    _regexp_constant = _re.compile('^.*-Cte$')

    def __init__(self, value):
        """Constant value."""
        self.value = value

    def read(self):
        """Return value."""
        return self.value

    @staticmethod
    def match(field):
        """Check if field is a constant."""
        return Constant._regexp_constant.match(field)


class ConstParameter:
    """Readable power supply parameter."""

    def __init__(self, pru_controller, device_id, param_id):
        """Init properties."""
        self.pru_controller = pru_controller
        self.device_id = device_id
        self.param_id = param_id
        self.value = None

    def read(self):
        """Read ps parameter from pru controller."""
        self.update()
        return self.value

    def update(self):
        """Update value."""
        self.value = self.pru_controller.read_parameters(
            self.device_id, self.param_id)
